searchData={"items":[{"type":"module","title":"Instructor","doc":"Check out our [Quickstart Guide](https://hexdocs.pm/instructor/quickstart.html) to get up and running with Instructor in minutes.\n\nInstructor provides structured prompting for LLMs. It is a spiritual port of the great [Instructor Python Library](https://github.com/jxnl/instructor) by [@jxnlco](https://twitter.com/jxnlco).\n\nInstructor allows you to get structured output out of an LLM using Ecto.  \nYou don't have to define any JSON schemas.\nYou can just use Ecto as you've always used it.  \nAnd since it's just ecto, you can provide change set validations that you can use to ensure that what you're getting back from the LLM is not only properly structured, but semantically correct.\n\nTo learn more about the philosophy behind Instructor and its motivations, check out this Elixir Denver Meetup talk:\n\n \n\n[![Instructor: Structured prompting for LLMs](assets/youtube-thumbnail.png)](https://www.youtube.com/watch?v=RABXu7zqnT0)\n\n \n\nWhile Instructor is designed to be used with OpenAI, it also supports every major AI lab and open source LLM inference server:\n\n- OpenAI\n- Anthropic\n- Groq\n- Ollama\n- Gemini\n- vLLM\n- llama.cpp\n\nAt its simplest, usage is pretty straightforward: \n\n1. Create an ecto schema, with a `@llm_doc` string that explains the schema definition to the LLM. \n2. Define a `validate_changeset/1` function on the schema, and use the `use Instructor` macro in order for Instructor to know about it.\n2. Make a call to `Instructor.chat_completion/1` with an instruction for the LLM to execute.\n\nYou can use the `max_retries` parameter to automatically, iteratively go back and forth with the LLM to try fixing validation errorswhen they occur.\n\n```elixir\nMix.install([:instructor])\n\ndefmodule SpamPrediction do\n  use Ecto.Schema\n  use Validator\n\n  @llm_doc \"\"\"","ref":"Instructor.html"},{"type":"module","title":"Field Descriptions: - Instructor","doc":"- class: Whether or not the email is spam.\n  - reason: A short, less than 10 word rationalization for the classification.\n  - score: A confidence score between 0.0 and 1.0 for the classification.\n  \"\"\"\n  @primary_key false\n  embedded_schema do\n    field(:class, Ecto.Enum, values: [:spam, :not_spam])\n    field(:reason, :string)\n    field(:score, :float)\n  end\n\n  @impl true\n  def validate_changeset(changeset) do\n    changeset\n    |> Ecto.Changeset.validate_number(:score,\n      greater_than_or_equal_to: 0.0,\n      less_than_or_equal_to: 1.0\n    )\n  end\nend\n\nis_spam? = fn text ->\n  Instructor.chat_completion(\n    model: \"gpt-4o-mini\",\n    response_model: SpamPrediction,\n    max_retries: 3,\n    messages: [\n      %{\n        role: \"user\",\n        content: \"\"\"\n        Your purpose is to classify customer support emails as either spam or not.\n        This is for a clothing retail business.\n        They sell all types of clothing.\n\n        Classify the following email: \n\n         \n          #{text}\n         \n        \"\"\"\n      }\n    ]\n  )\nend\n\nis_spam?.(\"Hello I am a Nigerian prince and I would like to send you money\")\n\n# => {:ok, %SpamPrediction{class: :spam, reason: \"Nigerian prince email scam\", score: 0.98}}\n```","ref":"Instructor.html#module-field-descriptions"},{"type":"function","title":"Instructor.cast_all/2","doc":"Casts all the parameters in the params map to the types defined in the types map.\nThis works both with Ecto Schemas and maps of Ecto types (see [Schemaless Ecto](https://hexdocs.pm/ecto/Ecto.Changeset.html#module-schemaless-changesets)).","ref":"Instructor.html#cast_all/2"},{"type":"function","title":"Examples - Instructor.cast_all/2","doc":"When using a full Ecto Schema\n\n    iex> Instructor.cast_all(%{\n    ...>   data: %Instructor.Demos.SpamPrediction{},\n    ...>   types: %{\n    ...>     class: :string,\n    ...>     score: :float\n    ...>   }\n    ...> }, %{\n    ...>   class: \"spam\",\n    ...>   score: 0.999\n    ...> })\n    %Ecto.Changeset{\n      action: nil,\n      changes: %{\n        class: \"spam\",\n        score: 0.999\n      },\n      errors: [],\n      data: %Instructor.Demos.SpamPrediction{\n        class: :spam,\n        score: 0.999\n      },\n      valid?: true\n    }\n\nWhen using a map of Ecto types\n\n    iex> Instructor.cast_all(%Instructor.Demo.SpamPrediction{}, %{\n    ...>   class: \"spam\",\n    ...>   score: 0.999\n    ...> })\n    %Ecto.Changeset{\n      action: nil,\n      changes: %{\n        class: \"spam\",\n        score: 0.999\n      },\n      errors: [],\n      data: %{\n        class: :spam,\n        score: 0.999\n      },\n      valid?: true\n    }\n\nand when using raw Ecto types,\n\n    iex> Instructor.cast_all({%{},%{name: :string}, %{\n    ...>   name: \"George Washington\"\n    ...> })\n    %Ecto.Changeset{\n      action: nil,\n      changes: %{\n        name: \"George Washington\",\n      },\n      errors: [],\n      data: %{\n        name: \"George Washington\",\n      },\n      valid?: true\n    }","ref":"Instructor.html#cast_all/2-examples"},{"type":"function","title":"Instructor.chat_completion/2","doc":"Create a new chat completion for the provided messages and parameters.\n\nThe parameters are passed directly to the LLM adapter.\nBy default they shadow the OpenAI API parameters.\nFor more information on the parameters, see the [OpenAI API docs](https://platform.openai.com/docs/api-reference/chat-completions/create).\n\nAdditionally, the following parameters are supported:\n\n  * `:adapter` - The adapter to use for chat completion. (defaults to the configured adapter, which defaults to `Instructor.Adapters.OpenAI`)\n  * `:response_model` - The Ecto schema to validate the response against, or a valid map of Ecto types (see [Schemaless Ecto](https://hexdocs.pm/ecto/Ecto.Changeset.html#module-schemaless-changesets)).\n  * `:stream` - Whether to stream the response or not. (defaults to `false`)\n  * `:validation_context` - The validation context to use when validating the response. (defaults to `%{}`)\n  * `:mode` - The mode to use when parsing the response, :tools, :json, :md_json (defaults to `:tools`), generally speaking you don't need to change this unless you are not using OpenAI.\n  * `:max_retries` - The maximum number of times to retry the LLM call if it fails, or does not pass validations.\n                     (defaults to `0`)","ref":"Instructor.html#chat_completion/2"},{"type":"function","title":"Examples - Instructor.chat_completion/2","doc":"iex> Instructor.chat_completion(\n    ...>   model: \"gpt-4o-mini\",\n    ...>   response_model: Instructor.Demos.SpamPrediction,\n    ...>   messages: [\n    ...>     %{\n    ...>       role: \"user\",\n    ...>       content: \"Classify the following text: Hello, I am a Nigerian prince and I would like to give you $1,000,000.\"\n    ...>     }\n    ...>   ])\n    {:ok,\n        %Instructor.Demos.SpamPrediction{\n            class: :spam\n            score: 0.999\n        }}\n\nWhen you're using Instructor in Streaming Mode, instead of returning back a tuple, it will return back a stream that emits tuples.\nThere are two main streaming modes available. array streaming and partial streaming.\n\nPartial streaming will emit the record multiple times until it's complete.\n\n    iex> Instructor.chat_completion(\n    ...>   model: \"gpt-4o-mini\",\n    ...>   response_model: {:partial, %{name: :string, birth_date: :date}}\n    ...>   messages: [\n    ...>     %{\n    ...>       role: \"user\",\n    ...>       content: \"Who is the first president of the United States?\"\n    ...>     }\n    ...>   ]) |> Enum.to_list()\n    [\n      {:partial, %{name: \"George Washington\"}},\n      {:partial, %{name: \"George Washington\", birth_date: ~D[1732-02-22]}},\n      {:ok, %{name: \"George Washington\", birth_date: ~D[1732-02-22]}}\n    ]\n\nWhereas with array streaming, you can ask the LLM to return multiple instances of your Ecto schema,\nand instructor will emit them one at a time as they arrive in complete form and validated.\n\n    iex> Instructor.chat_completion(\n    ...>   model: \"gpt-4o-mini\",\n    ...>   response_model: {:array, %{name: :string, birth_date: :date}}\n    ...>   messages: [\n    ...>     %{\n    ...>       role: \"user\",\n    ...>       content: \"Who are the first 5 presidents of the United States?\"\n    ...>     }\n    ...>   ]) |> Enum.to_list()\n\n    [\n      {:ok, %{name: \"George Washington\", birth_date: ~D[1732-02-22]}},\n      {:ok, %{name: \"John Adams\", birth_date: ~D[1735-10-30]}},\n      {:ok, %{name: \"Thomas Jefferson\", birth_date: ~D[1743-04-13]}},\n      {:ok, %{name: \"James Madison\", birth_date: ~D[1751-03-16]}},\n      {:ok, %{name: \"James Monroe\", birth_date: ~D[1758-04-28]}}\n    ]\n\nIf there's a validation error, it will return an error tuple with the change set describing the errors.\n\n    iex> Instructor.chat_completion(\n    ...>   model: \"gpt-4o-mini\",\n    ...>   response_model: Instructor.Demos.SpamPrediction,\n    ...>   messages: [\n    ...>     %{\n    ...>       role: \"user\",\n    ...>       content: \"Classify the following text: Hello, I am a Nigerian prince and I would like to give you $1,000,000.\"\n    ...>     }\n    ...>   ])\n    {:error,\n        %Ecto.Changeset{\n            changes: %{\n                class: \"foobar\",\n                score: -10.999\n            },\n            errors: [\n                class: {\"is invalid\", [type: :string, validation: :cast]}\n            ],\n            valid?: false\n        }}","ref":"Instructor.html#chat_completion/2-examples"},{"type":"type","title":"Instructor.stream/0","doc":"","ref":"Instructor.html#t:stream/0"},{"type":"behaviour","title":"Instructor.Adapter","doc":"Behavior for `Instructor.Adapter`.","ref":"Instructor.Adapter.html"},{"type":"callback","title":"Instructor.Adapter.chat_completion/2","doc":"","ref":"Instructor.Adapter.html#c:chat_completion/2"},{"type":"callback","title":"Instructor.Adapter.reask_messages/3","doc":"","ref":"Instructor.Adapter.html#c:reask_messages/3"},{"type":"type","title":"Instructor.Adapter.config/0","doc":"","ref":"Instructor.Adapter.html#t:config/0"},{"type":"type","title":"Instructor.Adapter.params/0","doc":"","ref":"Instructor.Adapter.html#t:params/0"},{"type":"type","title":"Instructor.Adapter.raw_response/0","doc":"","ref":"Instructor.Adapter.html#t:raw_response/0"},{"type":"type","title":"Instructor.Adapter.stream/0","doc":"","ref":"Instructor.Adapter.html#t:stream/0"},{"type":"module","title":"Instructor.Adapters.Anthropic","doc":"Anthropic adapter for Instructor.","ref":"Instructor.Adapters.Anthropic.html"},{"type":"module","title":"Configuration - Instructor.Adapters.Anthropic","doc":"```elixir\nconfig :instructor, adapter: Instructor.Adapters.Anthropic, anthropic: [\n  api_key: \"your_api_key\" # Will use ANTHROPIC_API_KEY environment variable if not provided\n]\n```\n\nor at runtime:\n\n```elixir\nInstructor.chat_completion(..., [\n  adapter: Instructor.Adapters.Anthropic,\n  api_key: \"your_api_key\" # Will use ANTHROPIC_API_KEY environment variable if not provided\n])\n```\n\nTo get an Anthropic API key, see [Anthropic](https://console.anthropic.com/settings/keys).","ref":"Instructor.Adapters.Anthropic.html#module-configuration"},{"type":"module","title":"Instructor.Adapters.Gemini","doc":"Adapter for Google Gemini.","ref":"Instructor.Adapters.Gemini.html"},{"type":"module","title":"Configuration - Instructor.Adapters.Gemini","doc":"```elixir\nconfig :instructor, adapter: Instructor.Adapters.Gemini, gemini: [\n  api_key: \"your_api_key\" # Will use GOOGLE_API_KEY environment variable if not provided\n]\n```\n\nor at runtime:\n\n```elixir\nInstructor.chat_completion(..., [\n  adapter: Instructor.Adapters.Gemini,\n  api_key: \"your_api_key\" # Will use GOOGLE_API_KEY environment variable if not provided\n])\n```\n\nTo get a Google API key, see [Google AI Studio](https://aistudio.google.com/apikey).","ref":"Instructor.Adapters.Gemini.html#module-configuration"},{"type":"function","title":"Instructor.Adapters.Gemini.chat_completion/2","doc":"Run a completion against Google's Gemini API\nAccepts OpenAI API arguments and converts to Gemini Args to perform the completion.\nDefaults to JSON mode within the Gemini API","ref":"Instructor.Adapters.Gemini.html#chat_completion/2"},{"type":"module","title":"Instructor.Adapters.Groq","doc":"Adapter for Groq Cloud API.  Using the OpenAI API compatible endpoint.","ref":"Instructor.Adapters.Groq.html"},{"type":"module","title":"Configuration - Instructor.Adapters.Groq","doc":"```elixir\nconfig :instructor, adapter: Instructor.Adapters.Groq, groq: [\n  api_key: \"your_api_key\" # Will use GROQ_API_KEY environment variable if not provided\n]\n```\n\nor at runtime:\n\n```elixir\nInstructor.chat_completion(..., [\n  adapter: Instructor.Adapters.Groq,\n  api_key: \"your_api_key\" # Will use GROQ_API_KEY environment variable if not provided\n])\n```\n\nFor more configurations, see Instructor.Adapters.OpenAI for more details as this adapter inherits functionality from it.\nTo get a Groq API key, see [Groq Cloud](https://groq.com/cloud).","ref":"Instructor.Adapters.Groq.html#module-configuration"},{"type":"module","title":"Instructor.Adapters.Llamacpp","doc":"Runs against the llama.cpp server. To be clear this calls the llamacpp specific\nendpoints, not the open-ai compliant ones.\n\nYou can read more about it here:\n  https://github.com/ggerganov/llama.cpp/tree/master/examples/server","ref":"Instructor.Adapters.Llamacpp.html"},{"type":"module","title":"Configuration - Instructor.Adapters.Llamacpp","doc":"```elixir\nconfig :instructor, adapter: Instructor.Adapters.Llamacpp, llamacpp: [\n  api_url: \"http://localhost:8080\" # (Optional) defaults to localhost:8080\n]\n```\n\nor at runtime:\n\n```elixir\nInstructor.chat_completion(..., [\n  adapter: Instructor.Adapters.Llamacpp,\n  api_url: \"http://localhost:8080\"\n])\n```","ref":"Instructor.Adapters.Llamacpp.html#module-configuration"},{"type":"function","title":"Instructor.Adapters.Llamacpp.chat_completion/2","doc":"Run a completion against the llama.cpp server, not the open-ai compliant one.\nThis gives you more specific control over the grammar, and the ability to\nprovide other parameters to the specific LLM invocation.\n\nYou can read more about the parameters here:\n  https://github.com/ggerganov/llama.cpp/tree/master/examples/server","ref":"Instructor.Adapters.Llamacpp.html#chat_completion/2"},{"type":"function","title":"Examples - Instructor.Adapters.Llamacpp.chat_completion/2","doc":"iex> Instructor.chat_completion(\n  ...>   model: \"llama3.1-8b-instruct\",\n  ...>   messages: [\n  ...>     %{ role: \"user\", content: \"Classify the following text: Hello I am a Nigerian prince and I would like to send you money!\" },\n  ...>   ],\n  ...>   response_model: response_model,\n  ...>   temperature: 0.5,\n  ...> )","ref":"Instructor.Adapters.Llamacpp.html#chat_completion/2-examples"},{"type":"module","title":"Instructor.Adapters.Ollama","doc":"Ollama adapter for Instructor.","ref":"Instructor.Adapters.Ollama.html"},{"type":"module","title":"Configuration - Instructor.Adapters.Ollama","doc":"```elixir\nconfig :instructor, adapter: Instructor.Adapters.Ollama, ollama: [\n  api_url: \"http://localhost:11434\" # (Optional) defaults to localhost:11434\n]\n```\n\nor at runtime:\n\n```elixir\nInstructor.chat_completion(..., [\n  adapter: Instructor.Adapters.Ollama,\n  api_url: \"http://localhost:11434\" # (Optional) defaults to localhost:11434\n])\n```\n\nTo get an Ollama API key, see [Ollama](https://ollama.com/api).","ref":"Instructor.Adapters.Ollama.html#module-configuration"},{"type":"module","title":"Instructor.Adapters.OpenAI","doc":"Documentation for `Instructor.Adapters.OpenAI`.","ref":"Instructor.Adapters.OpenAI.html"},{"type":"module","title":"Instructor.Adapters.VLLM","doc":"VLLM adapter for Instructor.","ref":"Instructor.Adapters.VLLM.html"},{"type":"module","title":"Configuration - Instructor.Adapters.VLLM","doc":"```elixir\nconfig :instructor, adapter: Instructor.Adapters.VLLM, vllm: [\n  api_url: \"http://localhost:8000\" # (Optional) defaults to localhost:8000\n]\n```\n\nor at runtime:\n\n```elixir\nInstructor.chat_completion(..., [\n  adapter: Instructor.Adapters.VLLM,\n  api_url: \"http://localhost:8000\" # (Optional) defaults to localhost:8000\n])\n```\n\nTo get a VLLM API key, see [VLLM](https://vllm.ai/).","ref":"Instructor.Adapters.VLLM.html#module-configuration"},{"type":"module","title":"Instructor.Adapters.XAI","doc":"Adapter for XAI API to access the Grok models.","ref":"Instructor.Adapters.XAI.html"},{"type":"module","title":"Configuration - Instructor.Adapters.XAI","doc":"```elixir\nconfig :instructor, adapter: Instructor.Adapters.XAI, xai: [\n  api_key: \"your_api_key\" # Will use XAI_API_KEY environment variable if not provided\n]\n```\n\nor at runtime:\n\n```elixir\nInstructor.chat_completion(..., [\n  adapter: Instructor.Adapters.XAI,\n  api_key: \"your_api_key\" # Will use XAI_API_KEY environment variable if not provided\n])\n```","ref":"Instructor.Adapters.XAI.html#module-configuration"},{"type":"behaviour","title":"Instructor.EctoType","doc":"Instructor.EctoType is a behaviour that lets your implement your own custom Ecto.Type\n  that works natively with Instructor.","ref":"Instructor.EctoType.html"},{"type":"behaviour","title":"Example - Instructor.EctoType","doc":"```elixir\n  defmodule MyCustomType do\n    use Ecto.Type\n    use Instructor.EctoType\n\n    # ... See `Ecto.Type` for implementation details\n\n    def to_json_schema() do\n      %{\n        type: \"string\",\n        format: \"email\"\n      }\n    end\n  end\n  ```","ref":"Instructor.EctoType.html#module-example"},{"type":"callback","title":"Instructor.EctoType.to_json_schema/0","doc":"","ref":"Instructor.EctoType.html#c:to_json_schema/0"},{"type":"module","title":"Instructor.Extras.ChainOfThought","doc":"","ref":"Instructor.Extras.ChainOfThought.html"},{"type":"function","title":"Instructor.Extras.ChainOfThought.chat_completion/2","doc":"","ref":"Instructor.Extras.ChainOfThought.html#chat_completion/2"},{"type":"module","title":"Instructor.Extras.ChainOfThought.ReasoningStep","doc":"","ref":"Instructor.Extras.ChainOfThought.ReasoningStep.html"},{"type":"function","title":"Instructor.Extras.ChainOfThought.ReasoningStep.__struct__/0","doc":"For each step, provide a title that describes what you're doing in that step, along with the content.\nDecide if you need another step or if you're ready to give the final answer.\nRespond in JSON format with 'title', 'content', and 'next_action' (either 'continue' or 'final_answer') keys.","ref":"Instructor.Extras.ChainOfThought.ReasoningStep.html#__struct__/0"},{"type":"module","title":"Instructor.JSONSchema","doc":"","ref":"Instructor.JSONSchema.html"},{"type":"function","title":"Instructor.JSONSchema.from_ecto_schema/1","doc":"Generates a JSON Schema from an Ecto schema.\n\n  Note: This will output a correct JSON Schema for the given Ecto schema, but\n  it will not necessarily be optimal, nor support all Ecto types.","ref":"Instructor.JSONSchema.html#from_ecto_schema/1"},{"type":"function","title":"Instructor.JSONSchema.traverse_and_update/3","doc":"Traverses a tree structure of maps and lists, allowing the user to update or remove elements.","ref":"Instructor.JSONSchema.html#traverse_and_update/3"},{"type":"function","title":"Parameters - Instructor.JSONSchema.traverse_and_update/3","doc":"- tree: The tree structure to traverse (can be a map, list, or any other type)\n  - fun: A function that takes either:\n    - Just the element if include_path: false (default)\n    - A tuple of {element, path} if include_path: true, where path is a list of keys to reach this element\n  The function should return either:\n    - An updated element\n    - nil to remove the element\n    - The original element if no changes are needed\n  - opts: Optional keyword list of options\n    - include_path: boolean, when true includes the path to each element in the callback (default: false)","ref":"Instructor.JSONSchema.html#traverse_and_update/3-parameters"},{"type":"function","title":"Returns - Instructor.JSONSchema.traverse_and_update/3","doc":"The updated tree structure","ref":"Instructor.JSONSchema.html#traverse_and_update/3-returns"},{"type":"module","title":"Instructor.JSONStreamParser","doc":"","ref":"Instructor.JSONStreamParser.html"},{"type":"function","title":"Instructor.JSONStreamParser.parse/1","doc":"","ref":"Instructor.JSONStreamParser.html#parse/1"},{"type":"behaviour","title":"Instructor.Validator","doc":"By default you'll get whatever OpenAI returns.\nThis behavior provides a hook for you to critique the response using standard ecto changesets validations.\nThis can be used in conjuction with the `:max_retries` parameter to `Instructor.chat_completion/1` to retry the completion until it passes your validation.","ref":"Instructor.Validator.html"},{"type":"behaviour","title":"Examples - Instructor.Validator","doc":"defmodule Instructor.Demos.SpamPrediction do\n    use Ecto.Schema\n\n    @primary_key false\n    schema \"spam_prediction\" do\n        field :class, :string\n        field :score, :float\n    end\n\n    @impl true\n    def validate_changeset(changeset) do\n        changeset\n        |> validate_number(:score, less_than_or_equal_to: 1.0)\n    end\n  end\n\n  iex> Instructor.chat_completion(\n  ...>   model: \"gpt-4o-mini\",\n  ...>   response_model: Instructor.Demos.SpamPrediction,\n  ...>   max_retries: 1,\n  ...>   messages: [\n  ...>     %{\n  ...>       role: \"user\",\n  ...>       content: \"Classify the following text: Hello, I am a Nigerian prince and I would like to give you $1,000,000.\"\n  ...>     }\n  ...>   ])\n  {:error, %Ecto.Changeset{\n      action: nil,\n      changes: %{},\n      errors: [\n          score: {\"is invalid\", [validation: :number, validation_opts: [less_than_or_equal_to: 1.0]]}\n      ],\n      data: %Instructor.Demos.SpamPrediction{\n          class: nil,\n          score: nil\n      },\n      valid?: false\n  }}","ref":"Instructor.Validator.html#module-examples"},{"type":"callback","title":"Instructor.Validator.validate_changeset/1","doc":"","ref":"Instructor.Validator.html#c:validate_changeset/1"},{"type":"callback","title":"Instructor.Validator.validate_changeset/2","doc":"","ref":"Instructor.Validator.html#c:validate_changeset/2"},{"type":"function","title":"Instructor.Validator.validate_with_llm/4","doc":"Validate a changeset field using a language model","ref":"Instructor.Validator.html#validate_with_llm/4"},{"type":"function","title":"Example - Instructor.Validator.validate_with_llm/4","doc":"defmodule QuestionAnswer do\n    use Ecto.Schema\n\n    @primary_key false\n    embedded_schema do\n        field :question, :string\n        field :answer, :string\n    end\n\n    @impl true\n    def validate_changeset(changeset) do\n        changeset\n        |> validate_with_llm(:answer, \"do not say anything objectionable\")\n    end\n  end","ref":"Instructor.Validator.html#validate_with_llm/4-example"},{"type":"extras","title":"Quickstart","doc":"<!-- livebook:{\"persist_outputs\":true} -->\n\n# Quickstart\n\n```elixir\nMix.install(\n  [\n    {:instructor, path: Path.expand(\"../\", __DIR__)}\n  ],\n  config: [\n    instructor: [\n      adapter: Instructor.Adapters.OpenAI,\n      openai: [api_key: System.fetch_env!(\"LB_OPENAI_API_KEY\")]\n    ]\n  ]\n)\n```","ref":"quickstart.html"},{"type":"extras","title":"Introduction - Quickstart","doc":"Instructor is a library to do structured prompting with OpenAI and open source LLMs. While the idea is pretty simple, through this and the other examples you'll realize how powerful a concept this is.\n\n**So first off, what is structure prompting?**\n\nWhat if the LLM returned data conforming to a complicated nested schema that your code knows how to work with? Well, that's structure prompting. It's a way of cohercing the LLM to producing it's response in a known format that your downstream code can handle. In the case of Instructor, we use Ecto to provide those schemas. Good old Ecto, something you're already familiar with.\n\nSo, without further ado, let's take define a schema and take it for a spin!\n\n```elixir\ndefmodule Politician do\n  use Ecto.Schema\n  use Instructor.Validator\n\n  @doc \"\"\"\n  A description of United States Politicians and the offices that they held,","ref":"quickstart.html#introduction"},{"type":"extras","title":"Fields: - Quickstart","doc":"- first_name: Their first name\n  - last_name: Their last name\n  - offices_held:\n    - office: The name of the political office held by the politician (in lowercase)\n    - from_date: When they entered office (YYYY-MM-DD)\n    - to_date: The date they left office, if relevant (YYYY-MM-DD or null).\n  \"\"\"\n  @primary_key false\n  embedded_schema do\n    field(:first_name, :string)\n    field(:last_name, :string)\n\n    embeds_many :offices_held, Office, primary_key: false do\n      field(:office, Ecto.Enum,\n        values: [:president, :vice_president, :governor, :congress, :senate]\n      )\n\n      field(:from_date, :date)\n      field(:to_date, :date)\n    end\n  end\nend\n```\n\n<!-- livebook:{\"output\":true} -->\n\n```\n{:module, Politician, <<70, 79, 82, 49, 0, 0, 17, ...>>,\n [__schema__: 1, __schema__: 1, __schema__: 1, __schema__: 1, __schema__: 2, __schema__: 2, ...]}\n```\n\nGreat, we have our schema describing politicans and the offices they held. Let's notice a few things that may stand out from regular Ecto usage. First, since there is no database backing the schema, it doesn't make sense to give it a primary_key. This also makes sense because there is no sensible value for the LLM to respond with.\n\nAlso we use a `@doc` on the schema. This isn't just for documentation purposes of the tutorial. Instructor will take any `@doc` tag and provide it to the LLM. Generally you'll want to use this to provide semantic descriptions of the fields and general context to the LLM to ensure you get the outputs you want. In our case we want to push the LLM to understand that we are only considering American politicians.\n\nSo, let's try asking the LLM to give us some politicians.\n\n```elixir\nInstructor.chat_completion(\n  model: \"gpt-4o-mini\",\n  response_model: Politician,\n  messages: [\n    %{\n      role: \"user\",\n      content:\n        \"Who won the American 2020 election and what offices have they held over their career?\"\n    }\n  ]\n)\n```\n\n<!-- livebook:{\"output\":true} -->\n\n```\n{:ok,\n %Politician{\n   first_name: \"Joe\",\n   last_name: \"Biden\",\n   offices_held: [\n     %Politician.Office{office: :president, from_date: ~D[2021-01-20], to_date: nil},\n     %Politician.Office{office: :vice_president, from_date: ~D[2009-01-20], to_date: ~D[2017-01-20]},\n     %Politician.Office{office: :senate, from_date: ~D[1973-01-03], to_date: ~D[2009-01-15]}\n   ]\n }}\n```\n\nAmazing, right? Using nothing more than one of the top libraries in Elixir, Ecto, we were able to get structured output from our LLM. The data returned is ready to be processed by our regular Elixir code. Instructor supports all field types that you can express in Ecto, including embedded and associated schemas.\n\nIt's almost as if the LLM inputted the data into a Phoenix Form. All the utilities that you use to process that kind of data, you can use to process the outputs of Instructor.\n\nOne of the superpowers of this is that since we're just using changesets under the hood, you can use the same validations that you would use elsewhere in your app. Let's look at that in the next section.\n\n<!-- livebook:{\"branch_parent_index\":0} -->","ref":"quickstart.html#fields"},{"type":"extras","title":"Validations - Quickstart","doc":"Instructor provides a lightweight behavior where you can define a callback function that we will call to validate the data returned by the LLM using Ecto changesets. There is nothing fancy to this API. It's just a changeset in and a changeset out.\n\n```elixir\ndefmodule NumberSeries do\n  use Ecto.Schema\n  use Instructor.Validator\n\n  @primary_key false\n  embedded_schema do\n    field(:series, {:array, :integer})\n  end\n\n  @impl true\n  def validate_changeset(changeset) do\n    changeset\n    |> Ecto.Changeset.validate_length(:series, min: 10)\n    |> Ecto.Changeset.validate_change(:series, fn\n      field, values ->\n        if Enum.sum(values) |> rem(2) == 0 do\n          []\n        else\n          [{field, \"The sum of the series must be even\"}]\n        end\n    end)\n  end\nend\n```\n\n<!-- livebook:{\"output\":true} -->\n\n```\n{:module, NumberSeries, <<70, 79, 82, 49, 0, 0, 18, ...>>, {:validate_changeset, 1}}\n```\n\nIn this albeit contrived example, we're going to get the LLM to return a series of numbers and validate whether it has at least 10 numbers and that the sum of the series is even.\n\nWhen we ask for fewer than ten numbers, Instructor will return an error tuple with a change set that is invalid.\n\n```elixir\n{:error, changeset} =\n  Instructor.chat_completion(\n    model: \"gpt-4o-mini\",\n    response_model: NumberSeries,\n    messages: [\n      %{role: \"user\", content: \"Give me the first 5 integers\"}\n    ]\n  )\n\n# Render our the errors down to strings.\nerrors =\n  Ecto.Changeset.traverse_errors(changeset, fn {msg, opts} ->\n    Regex.replace(~r\"%{(\\w+)}\", msg, fn _, key ->\n      opts |> Keyword.get(String.to_existing_atom(key), key) |> to_string()\n    end)\n  end)\n\n{changeset.changes, errors}\n```\n\n<!-- livebook:{\"output\":true} -->\n\n```\n{%{series: [1, 2, 3, 4, 5]},\n %{series: [\"The sum of the series must be even\", \"should have at least 10 item(s)\"]}}\n```\n\nNow the beauty of this is that since we have human readable errors from our validations, we can just turn around and pass those back into the LLM to get it to fix its own errors.\n\nInstructor provides a convenience parameter, `max_retries` for you in the initial call which will retry against the validations up to n times.\n\n```elixir\nInstructor.chat_completion(\n  model: \"gpt-4o-mini\",\n  response_model: NumberSeries,\n  max_retries: 10,\n  messages: [\n    %{role: \"user\", content: \"Give some random integers\"}\n  ]\n)\n```\n\n<!-- livebook:{\"output\":true} -->\n\n```\n\n10:30:03.764 [debug] Retrying LLM call for NumberSeries:\n\n \"series - The sum of the series must be even\\nseries - should have at least 10 item(s)\"\n\n10:30:04.794 [debug] Retrying LLM call for NumberSeries:\n\n \"series - The sum of the series must be even\"\n\n```\n\n<!-- livebook:{\"output\":true} -->\n\n```\n{:ok,\n %NumberSeries{series: [1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 20]}}\n```\n\nHere we demonstrated using regular Lixar code to validate the outputs of an LLM, but we don't have to stop there. We can actually use the LLM to validate the outputs of the LLM.\n\nIn Instructor, we have provided a custom Ecto Changset validator called `validate_with_llm`. \nUnder the hood it just uses an instructor itself to check whether the field matches some condition that you have defined in plain text.\n\n```elixir\ndefmodule QuestionAnswer do\n  use Ecto.Schema\n  use Instructor.Validator\n\n  @primary_key false\n  embedded_schema do\n    field(:question, :string)\n    field(:answer, :string)\n  end\n\n  @impl true\n  def validate_changeset(changeset) do\n    changeset\n    |> validate_with_llm(:answer, \"Do not say anything objectionable\")\n  end\nend\n```\n\n<!-- livebook:{\"output\":true} -->\n\n```\n{:module, QuestionAnswer, <<70, 79, 82, 49, 0, 0, 16, ...>>, {:validate_changeset, 1}}\n```\n\n```elixir\n%QuestionAnswer{}\n|> Instructor.cast_all(%{\n  question: \"What is the meaning of life?\",\n  answer: \"Sex, drugs, and rock'n roll\"\n})\n|> QuestionAnswer.validate_changeset()\n```\n\n<!-- livebook:{\"output\":true} -->\n\n```\n#Ecto.Changeset ,\n  valid?: false\n>\n```\n\n<!-- livebook:{\"branch_parent_index\":0} -->","ref":"quickstart.html#validations"},{"type":"extras","title":"Record Streaming - Quickstart","doc":"Now if you've used chatGPT's web interface, you know that these LLMs can stream responses one token at a time. You can imagine that this is pretty easy to implement in code as you just reduce across the stream accumulating the value and appending to the log. But doing this when you're using OpenAI's function calls, and where the data is structured in JSON, it's not trivial to implement streaming.\n\nLuckily we've done that work for you and we support two types of streaming in Instructor. The first is **record streaming** and the second is **partial streaming**.\n\nFirst let's take a look at record streaming.\n\nRecord streaming is useful when you're asking the LLM for something that is naturally represented as an array of records. In this mode, we will instead of returning the full array, we'll return a stream that will emit each record once it's been completely streamed to the client, but before the next records tokens have arrived.\n\nFor example, let's take our presidents example from earlier, and we can instead ask for the first 5 presidents of the United States streaming each result as they come in.\n\n```elixir\npresidents_stream =\n  Instructor.chat_completion(\n    model: \"gpt-4o-mini\",\n    stream: true,\n    response_model: {:array, Politician},\n    messages: [\n      %{role: \"user\", content: \"Who were the first 5 presidents of the United States?\"}\n    ]\n  )\n```\n\n<!-- livebook:{\"output\":true} -->\n\n```\n#Stream<[\n  enum: #Function<60.53678557/2 in Stream.transform/3>,\n  funs: [#Function<48.53678557/1 in Stream.map/2>]\n]>\n```\n\nAs you can see, instead of returning the result, we return a stream which can be run to emit each of the presidents.\n\n```elixir\npresidents_stream\n|> Stream.each(fn\n  {:ok, politician} -> IO.inspect(politician)\n  {:error, changeset} -> IO.inspect(changeset)\nend)\n|> Stream.run()\n```\n\n<!-- livebook:{\"output\":true} -->\n\n```\n%Politician{\n  first_name: \"George\",\n  last_name: \"Washington\",\n  offices_held: [\n    %Politician.Office{\n      office: :president,\n      from_date: ~D[1789-04-30],\n      to_date: ~D[1797-03-04]\n    }\n  ]\n}\n%Politician{\n  first_name: \"John\",\n  last_name: \"Adams\",\n  offices_held: [\n    %Politician.Office{\n      office: :president,\n      from_date: ~D[1797-03-04],\n      to_date: ~D[1801-03-04]\n    }\n  ]\n}\n%Politician{\n  first_name: \"Thomas\",\n  last_name: \"Jefferson\",\n  offices_held: [\n    %Politician.Office{\n      office: :president,\n      from_date: ~D[1801-03-04],\n      to_date: ~D[1809-03-04]\n    }\n  ]\n}\n%Politician{\n  first_name: \"James\",\n  last_name: \"Madison\",\n  offices_held: [\n    %Politician.Office{\n      office: :president,\n      from_date: ~D[1809-03-04],\n      to_date: ~D[1817-03-04]\n    }\n  ]\n}\n%Politician{\n  first_name: \"James\",\n  last_name: \"Monroe\",\n  offices_held: [\n    %Politician.Office{\n      office: :president,\n      from_date: ~D[1817-03-04],\n      to_date: ~D[1825-03-04]\n    }\n  ]\n}\n```\n\n<!-- livebook:{\"output\":true} -->\n\n```\n:ok\n```\n\nAn important thing to note here is that we're running the validations independently for each value in the array.  That's why the values in the stream are either `{:ok, Ecto.Schema.t()}` or `{:error, Ecto.Changeset.t()}`.\n\nAs a result, it's unclear how we can automatically do retries to fix validation errors. And therefore, when in streaming mode, it is the responsibility of the user to retry when validation errors occur. (We may revisit this decision in the future)\n\n<!-- livebook:{\"branch_parent_index\":0} -->","ref":"quickstart.html#record-streaming"},{"type":"extras","title":"Partial Streaming - Quickstart","doc":"The other streaming mode that we have an instructor is called partial streaming. In this mode, you can get back a stream that will emit the record multiple times with the fields updating as they arrive. This can be used with a schema or an array of schemas. Both are demonstrated below.\n\nThis is useful in UI applications where you want to show instant feedback to the user about what data is showing up when without giving just some indeterminant loading spinner.\n\n```elixir\nInstructor.chat_completion(\n  model: \"gpt-4o-mini\",\n  stream: true,\n  response_model: {:partial, Politician},\n  messages: [\n    %{role: \"user\", content: \"Who is the first president of the United States?\"}\n  ]\n)\n|> Stream.each(fn\n  {:partial, politician} -> IO.puts(\"[Partial]: #{inspect(politician)}\")\n  {:ok, politician} -> IO.puts(\"[Final]: #{inspect(politician)}\")\nend)\n|> Stream.run()\n```\n\n<!-- livebook:{\"output\":true} -->\n\n```\n[Partial]: %Politician{first_name: nil, last_name: nil, offices_held: []}\n[Partial]: %Politician{first_name: nil, last_name: nil, offices_held: []}\n[Partial]: %Politician{first_name: \"George\", last_name: nil, offices_held: []}\n[Partial]: %Politician{first_name: \"George\", last_name: \"Washington\", offices_held: []}\n[Partial]: %Politician{first_name: \"George\", last_name: \"Washington\", offices_held: []}\n[Partial]: %Politician{first_name: \"George\", last_name: \"Washington\", offices_held: [%Politician.Office{office: nil, from_date: nil, to_date: nil}]}\n[Partial]: %Politician{first_name: \"George\", last_name: \"Washington\", offices_held: [%Politician.Office{office: nil, from_date: ~D[1789-04-30], to_date: nil}]}\n[Partial]: %Politician{first_name: \"George\", last_name: \"Washington\", offices_held: [%Politician.Office{office: :president, from_date: ~D[1789-04-30], to_date: nil}]}\n[Partial]: %Politician{first_name: \"George\", last_name: \"Washington\", offices_held: [%Politician.Office{office: :president, from_date: ~D[1789-04-30], to_date: ~D[1797-03-04]}]}\n[Final]: %Politician{first_name: \"George\", last_name: \"Washington\", offices_held: [%Politician.Office{office: :president, from_date: ~D[1789-04-30], to_date: ~D[1797-03-04]}]}\n```\n\n<!-- livebook:{\"output\":true} -->\n\n```\n:ok\n```\n\nThere is an important difference in this mode. Since your validations will be defined on the entirety of the object, it doesn't make sense to call the validate function until the entire record has been streamed in.\n\nTherefore, we introduce a new output tuple in the stream compared to regular record streaming. The value can be `{:partial, Ecto.Schema.t()}`, and then on the last emit of the stream it can be `{:error, Ecto.Changeset.t()}`, or `{:ok, Ecto.Schema.t()}`\n\nLike record streaming, however, using `max_retries` with this streaming mode does nothing and throws an error. (We may revisit this in the future when it's clear what such a behavior should do)\n\n<!-- livebook:{\"branch_parent_index\":0} -->","ref":"quickstart.html#partial-streaming"},{"type":"extras","title":"Custom Ecto Types - Quickstart","doc":"Instructor supports all the Ecto types out of the box, but sometimes you need more. And that's why Instructor provides a behavior that you can implement on your own custom Ecto types. All you have to do is implement `to_json_schema/0`.\n\nWhatever you return from this function will be put as the field type. See the [JSONSchema Specification](https://json-schema.org/understanding-json-schema/reference) for more information on what you can put here. Typically you'll see people put `description`, `type`, and maybe `format`.\n\n```elixir\ndefmodule EctoURI do\n  use Ecto.Type\n  use Instructor.EctoType\n\n  def type, do: :map\n\n  # This is it, the rest is for implementing a regular old ecto type.\n  def to_json_schema() do\n    %{\n      type: \"string\",\n      description: \"A valid URL\"\n    }\n  end\n\n  def cast(uri) when is_binary(uri) do\n    {:ok, URI.parse(uri)}\n  end\n\n  def cast(%URI{} = uri), do: {:ok, uri}\n  def cast(_), do: :error\n\n  def load(data) when is_map(data) do\n    data =\n      for {key, val} <- data do\n        {String.to_existing_atom(key), val}\n      end\n\n    {:ok, struct!(URI, data)}\n  end\n\n  def dump(%URI{} = uri), do: {:ok, Map.from_struct(uri)}\n  def dump(_), do: :error\nend\n```\n\n<!-- livebook:{\"output\":true} -->\n\n```\n{:module, EctoURI, <<70, 79, 82, 49, 0, 0, 14, ...>>, {:dump, 1}}\n```\n\n```elixir\nInstructor.chat_completion(\n  model: \"gpt-4o-mini\",\n  response_model: %{url: EctoURI},\n  messages: [\n    %{role: \"user\", content: \"Give me the URL for Google\"}\n  ]\n)\n```\n\n<!-- livebook:{\"output\":true} -->\n\n```\n{:ok,\n %{\n   url: %URI{\n     scheme: \"https\",\n     authority: \"www.google.com\",\n     userinfo: nil,\n     host: \"www.google.com\",\n     port: 443,\n     path: nil,\n     query: nil,\n     fragment: nil\n   }\n }}\n```\n\nAnd just like that, you can extend Instructor to get the LLM to return whatever you want.\n\n<!-- livebook:{\"offset\":16817,\"stamp\":{\"token\":\"XCP.5TJEWZKV766d3PqSlapF6zfDrzPBF2D8YZUn1BV9Fgs73D-kM1ZIpv6KHUXjgokTe8_FejNl25O7e0oAjBtqUTEnpAzkOGBlPiJirCXs-FR86ffK3qxBXWE\",\"version\":2}} -->","ref":"quickstart.html#custom-ecto-types"},{"type":"extras","title":"Philosophy & Motivations","doc":"# Philosophy & Motivations\n\nInstructor is simple library that allows you to maximally leverage large language models (LLMs).\nIt acts as your bridge between [Karpathy's Software 1.0 and Software 2.0](https://karpathy.medium.com/software-2-0-a64152b37c35).\n\nSoftware 1.0 is the world we're all used to. \nIt's datastructures, databases, algorithms, code.\n\nSoftware 2.0 on the other hand is neural networks.\nThey're models to problems for which it's easier to gather examples of desired behavior than describe how to produce it.","ref":"philosophy.html"},{"type":"extras","title":"A New Modality - Philosophy & Motivations","doc":"Software 2.0, as it exists today operates in the modality of text, images, and audio.\nAll modalities that have been traditionally very difficult to deal with in Software 1.0.\n\nWe now have these things called LLMs.\nFor the first time we can solve non-trivial `f(text) -> text` problems.\n\n```\n\n  llm(\"Recipe for a banana smoothie\") -> \"\"\"\n    A banana smoothie is a delicious and nutritious treat that's quite easy to make. Here's a simple recipe for you:\n\n    Ingredients:\n\n    2 ripe bananas\n    1 cup milk (you can use almond, soy, or cow's milk)\n    1/2 cup Greek yogurt (optional, for extra creaminess)\n    1-2 tablespoons honey or maple syrup (adjust according to taste)\n    A pinch of cinnamon (optional)\n    Ice cubes (optional, for a colder smoothie)\n    Instructions:\n\n    Peel the bananas and place them in a blender.\n    Add the milk, Greek yogurt (if using), honey/maple syrup, and cinnamon.\n    Add a few ice cubes if you prefer a colder smoothie.\n    Blend until smooth and creamy. If the smoothie is too thick, you can add a little more milk to reach your desired consistency.\n    Taste and adjust the sweetness if necessary.\n    Pour into glasses and serve immediately.\n\n    Enjoy your banana smoothie!\n  \"\"\"\n\n```\n\nHowever from a Software 1.0 perspective, this isn't terribly useful.\nRemember, Software 1.0 doesn't have robust, non trivial semantic text processing.\nWe can pretty easily create the function's arguments, but are you sure you can robustly parse the output? Unlikely.","ref":"philosophy.html#a-new-modality"},{"type":"extras","title":"LLMs aren't only for chatbots - Philosophy & Motivations","doc":"Over the past year you've probably noticed that every LLM company produces some variation of a chatbot -\nChat with GPT, Chat with your PDF, Chat with your AI Girlfriend.\n\nThis makes sense, when you think about it.\nWe're really good at creating strings in Software 1.0 from structured data, but we're terrible at creating structured data from strings.\nNor are we any good any good at interpreting the semantic meaning of strings.\n\nTherefore, it only makes sense to design experiences where we present them to the user for interpretation.\nAlas, our ubiquitous chatbot interface.\n\nBut it doesn't have to be that way. \nWhat if we can get our LLMs to respond with structured data?\nThat'd give us the all our domains and ranges covered,\n\n```\n  f(text) -> text\n  f(data_structure |> to_string) -> text\n  f(data_structure |> to_string) -> data_structure\n```\n\nWith such a capability, you would have full bidirectional interoperability between software 1.0 and software 2.0. \nThat's what Instructor provides.","ref":"philosophy.html#llms-aren-t-only-for-chatbots"},{"type":"extras","title":"cast_all\\2, It's just Ecto baby - Philosophy & Motivations","doc":"Now to do this we need to give the LLM a schema that it has to conform to in its output text. \nIf we can guarantee that it outputs JSON that matches this schema, we can treat it as any other user data we're used to in Elixir.\n\nThat's why, we figure, let's just build this on top of Ecto. It's already in your app after all.\nYou already know how to use it.\nIt already provides a robust validations API.\nLet's just treat the LLM as if it's one of your users, submitting data through a form.\n\n```elixir\ndefmodule Recipe do\n  use Ecto.Schema\n\n  @doc \"\"\"\n    Our AI generated delicious recipe.\n  \"\"\"\n  @primary_key false\n  embedded_schema do \n    field :title, :string\n    field :cook_time, :integer\n    field :steps, {:array, :string}\n    embeds_many :ingredients, Ingredients, primary_key: false do\n      field :name, :string\n      field :quantity, :decimal\n      field :unit, :string\n    end\n  end\nend\n```\n\nFrom this instructor takes your ecto schema, converts it into a JSON schema, and passes it on to the LLM.\nIt uses clever techniques like function calling and BNF grammar sampling to ensure that we get back a result that is exactly matching our Ecto schema.\nAt which point it's all just Ecto code.\nWe cast the fields, we do the validations, and we return the result back to you.\n\n```elixir\nInstructor.chat_completion(\n  model: \"gpt-4o-mini\",\n  response_model: Recipe,\n  messages: [\n    %{ role: \"user\", content: \"Give me a recipe for a banana smoothie\" }\n  ]\n)\n\n# => {:ok, %Recipe{title: \"Banana smoothie\", cook_time: 15, steps: [...]}\n```\n\nInstructor makes Ecto is our interchange format between Software 1.0 and 2.0.\nPerfect because Ecto is a mainstay in the ecosystem and compatible with so many other libraries.","ref":"philosophy.html#cast_all-2-it-s-just-ecto-baby"},{"type":"extras","title":"Used Anywhere where Ecto is used - Philosophy & Motivations","doc":"Now this is a neat capability, but you may be thinking, where can I use this?  \nThis is Virgin Territory. We've never had this capability. It is a new modality in computing.\nIt will require a bit of imagination.\nA good heuristic to start with is, \"wherever I use Ecto, I can use Instructor.\" \n\n```\n  user -> instructor -> code\n  code -> instructor -> code\n  code -> instructor -> user\n```\n\nThere is a 40Gb LLM file sitting on Huggingface that encodes the entirety of human knowledge, you now have a query interface for it.","ref":"philosophy.html#used-anywhere-where-ecto-is-used"},{"type":"extras","title":"Where to go from here? - Philosophy & Motivations","doc":"First, go check out the [Introduction to Instructor Tutorial](#) where we go over the user guide of how to get up and running with instructor.\nThen, checkout the [cookbooks](#) to see various techniques of how to best leverage instructor in common usecases.","ref":"philosophy.html#where-to-go-from-here"},{"type":"extras","title":"Local Instructor w/ llama.cpp","doc":"<!-- livebook:{\"persist_outputs\":true} -->\n\n# Local Instructor w/ llama.cpp\n\n```elixir\nMix.install(\n  [\n    {:instructor, path: Path.expand(\"../../\", __DIR__)},\n    {:kino_shell, \"~> 0.1.2\"}\n  ],\n  config: [\n    instructor: [\n      adapter: Instructor.Adapters.Llamacpp,\n      llamacpp: [\n        chat_template: :mistral_instruct\n      ]\n    ]\n  ]\n)\n```","ref":"llama-cpp.html"},{"type":"extras","title":"Setting up llama.cpp - Local Instructor w/ llama.cpp","doc":"[llama.cpp](https://github.com/ggerganov/llama.cpp) is a great way to run models locally. Head on over to the repo and install it on your system.\n\nNext, we'll need to download a GGUF compatible model to run with llama.cpp. As of today, I recommend using qwen-2.5-7b. It's a great model that is small enough to run locally.\n\n> **A note on quantization:** When you go search for GGUF models you'll see a lot of suffixes like `Q4_K_M` and `F8`. These are just different compression techniques, called quantization, that allow the model to take up dramatically less memory at the cost of some accuracy. There are many different methods of quantization that have different performance tradeoffs. However, it's generally recommended to run the largest model you can fit into your GPUs VRAM. Going over FP8 is generally unnecessary and at that point, you should be considering models with a larger number of parameters.\n\nTo start the llama server, run `llama-server --port 8080 -ngl 999 -hf Qwen/Qwen2.5-7B-Instruct-GGUF`. This will automatically download the model weights and start a server to run them. `-ngl 999` is just a flag to tell llama how many layers of the neural network to offload to the GPU. 999, is effectively saying, run the entire model on the GPU.\n\nThen with that running in the background. You can use Instructor as you normally would!\n\nThere are three ways to configure Instructor to use llama.cpp,\n\n1. via `Mix.install([...], [instructor: [adapter: Instructor.Adapters.Llamacpp, llamacpp: [...]]])`\n2. via `config :instructor, adapter: Instructor.Adapters.Ollama, llamacpp: [...]`\n3. At runtime via `Instructor.chat_completion(..., config)`\n\nFor brevity, in this livebook, we'll configure it at runtime\n\n```elixir\nconfig = [\n  adapter: Instructor.Adapters.Llamacpp\n]\n\ndefmodule President do\n  use Ecto.Schema\n\n  @primary_key false\n  embedded_schema do\n    field(:first_name, :string)\n    field(:last_name, :string)\n    field(:entered_office_date, :date)\n  end\nend\n\nInstructor.chat_completion(\n  [\n    response_model: President,\n    mode: :json_schema,\n    messages: [\n      %{role: \"user\", content: \"Who was the first president of the United States?\"}\n    ]\n  ],\n  config\n)\n```\n\n<!-- livebook:{\"output\":true} -->\n\n```\n{:ok,\n %President{first_name: \"George\", last_name: \"Washington\", entered_office_date: ~D[1789-04-01]}}\n```\n\nAnd there you have it. You're running Instructor against a locally running large language model.  At zero incremental cost to you.","ref":"llama-cpp.html#setting-up-llama-cpp"},{"type":"extras","title":"Local Instructor w/ Ollama","doc":"# Local Instructor w/ Ollama\n\n```elixir\nMix.install(\n  [\n    {:instructor, path: Path.expand(\"../../\", __DIR__)}\n  ]\n)\n```","ref":"ollama.html"},{"type":"extras","title":"Introduction - Local Instructor w/ Ollama","doc":"Before running the code below, please ensure the following:\n\n1. **Pull the Required Model**:  \n   Run the following command to pull the necessary model image:\n\n   ```bash\n   ollama pull qwen2.5:7b\n   ```\n\n   This command downloads the model `qwen2.5:7b` so that it can be used by the Instructor adapter.\n\n2. **Start the server**:  \n   Run the following command to start the local ollama server:\n\n   ```bash\n   ollama serve\n   ```\n\nOnce you have the server running, you can call it with Instructor using the Ollama Adapter.\n\nThere are three ways to configure Instructor to use Ollama,\n\n1. via `Mix.install([...], [instructor: [adapter: Instructor.Adapters.Ollama, ollama: [...]]])`\n2. via `config :instructor, adapter: Instructor.Adapters.Ollama, ollama: [...]`\n3. At runtime via `Instructor.chat_completion(..., config)`\n\nFor brevity, in this livebook, we'll configure it at runtime\n\n```elixir\nconfig = [\n  adapter: Instructor.Adapters.Ollama\n]\n\ndefmodule President do\n  use Ecto.Schema\n  use Instructor\n\n  @primary_key false\n  embedded_schema do\n    field(:first_name, :string)\n    field(:last_name, :string)\n    field(:entered_office_date, :date)\n  end\nend\n\nInstructor.chat_completion(\n  [\n    model: \"qwen2.5:7b\",\n    mode: :json,\n    response_model: President,\n    messages: [\n      %{role: \"user\", content: \"Who was the first president of the United States?\"}\n    ]\n  ],\n  config\n)\n```\n\nThere are three ways to configure Instructor to use Ollama,\n\n1. via `Mix.install([...], config)`\n2. via `config :instructor, adapter: :ollama, ollama: config`\n3. At runtime via `Instructor.chat_completion(..., config)`\n\nFor brevity, in this livebook, we'll configure it at runtime","ref":"ollama.html#introduction"},{"type":"extras","title":"Local Instructor w/ vLLM","doc":"# Local Instructor w/ vLLM\n\n```elixir\nMix.install(\n  [\n    {:instructor, path: Path.expand(\"../../\", __DIR__)},\n  ]\n)\n```","ref":"vllm.html"},{"type":"extras","title":"Introduction - Local Instructor w/ vLLM","doc":"When it comes to local inference, there are three main tools that people use: olama, llama.cpp, and in professional settings, vllm.  There are obviously more, but these are the main ones that people use.  vLLM is a high performance inference server that can handle many concurrent requests in parallel.  It uses this feature called grouped attention to make sure it can utilize the GPU's VRAM very efficiently.  On an RTX 4090, you can expect the LLM to push ~3000 tokens/sec on a 7B parameter model.\n\nvLLM is a great option if you want to host a local LLM inference server on an old gaming machine.  Or, in a corporate environment, you can host this on a GPU-optimized EC2 instance for completely private inference.\n\nTo install vLLM, head on over to the docks and run through their [quick start guide](https://docs.vllm.ai/en/latest/getting_started/quickstart.html).\n\nOnce installed, you can start up an OpenAI compliant inference server by running the following command,\n\n```\n$ vllm serve Qwen/Qwen2.5-1.5B-Instruct\n```\n\nThis will download the model and start up your inference server. Instructor will plug in seamlessly when using the vLLM adapter.\n\nThere are three ways to configure Instructor to use vLLM,\n\n1. via `Mix.install([...], [instructor: [adapter: Instructor.Adapters.VLLM, vllm: [...]]])`\n2. via `config :instructor, adapter: Instructor.Adapters.VLLM, vllm: [...]`\n3. At runtime via `Instructor.chat_completion(..., config)`\n\nFor brevity, in this livebook, we'll configure it at runtime\n\n```elixir\nconfig = [\n  adapter: Instructor.Adapters.VLLM\n]\n\ndefmodule President do\n  use Ecto.Schema\n  use Instructor\n\n  @primary_key false\n  embedded_schema do\n    field(:first_name, :string)\n    field(:last_name, :string)\n    field(:entered_office_date, :date)\n  end\nend\n\nInstructor.chat_completion(\n  [\n    model: \"Qwen/Qwen2.5-1.5B-Instruct\",\n    mode: :json_schema,\n    response_model: President,\n    messages: [\n      %{role: \"user\", content: \"Who was the first president of the United States?\"}\n    ]\n  ],\n  config\n)\n```","ref":"vllm.html#introduction"},{"type":"extras","title":"Gemini","doc":"# Gemini\n\n```elixir\nMix.install(\n  [\n    {:instructor, path: Path.expand(\"../../\", __DIR__)},\n  ]\n)\n```","ref":"gemini.html"},{"type":"extras","title":"Introduction - Gemini","doc":"Google is in the LLM game, i'm sure you've heard. Currently not many people use them, however that's not for lack of quality or for absurding pricing. Actually, they're quite cheap and incredibly good for the price. The main reason why people don't use them is that they don't use an OpenAI-compliant API. Therefore most of the libraries don't support it. Not instructor, no. We support it. We have to take some liberties to ensure that the API conforms, but it works and you can be up and running before you know it.\n\nFirst, head on over to the [Google AI Studio](https://aistudio.google.com/apikey) to get an API key.\n\nThere are three ways to configure Instructor to use Gemini.\n\n1. via `Mix.install([...], [instructor: [adapter: Instructor.Adapters.Gemini, gemini: [...]]])`\n2. via `config :instructor, adapter: Instructor.Adapters.Gemini, gemini: [...]`\n3. At runtime via `Instructor.chat_completion(..., config)`\n\nFor brevity, in this livebook, we'll configure it at runtime.\n\nWe recommend that you use Instructor with `mode: :json_schema`.\n\n```elixir\nconfig = [\n  adapter: Instructor.Adapters.Gemini,\n  api_key: System.fetch_env!(\"LB_GEMINI_API_KEY\")\n]\n\ndefmodule President do\n  use Ecto.Schema\n\n  @primary_key false\n  embedded_schema do\n    field(:first_name, :string)\n    field(:last_name, :string)\n    field(:entered_office_date, :date)\n  end\nend\n\nInstructor.chat_completion(\n  [\n    model: \"gemini-2.0-flash\",\n    mode: :json_schema,\n    response_model: President,\n    messages: [\n      %{role: \"user\", content: \"Who was the first president of the United States?\"}\n    ]\n  ],\n  config\n)\n```\n\nAnd there you have it. Instructor with Google Gemini.\n\n<!-- livebook:{\"offset\":1778,\"stamp\":{\"token\":\"XCP.wUwkJdv008d3KLA939_kGrHTvgXTyGpSyRIYvpDoM0LuMzuOZsGoifUHc18BeQOzlzihEXnO9FCqcp5tO2_4Q5NiXuAN3YRczFrjaU-Qg9yHeG-9FtM6sk4\",\"version\":2}} -->","ref":"gemini.html#introduction"},{"type":"extras","title":"Anthropic","doc":"# Anthropic\n\n```elixir\nMix.install(\n  [\n    {:instructor, path: Path.expand(\"../../\", __DIR__)},\n  ]\n)\n```","ref":"anthropic.html"},{"type":"extras","title":"Introduction - Anthropic","doc":"Anthropic is a leading AI lab that competes with both OpenAI and Google.  Their models rank among the top, though their pricing is a little more expensive.  Lucky for you, we support them out of the box.  All you need to do is create an account with [Anthropic](https://console.anthropic.com/settings/keys) and get an API key.\n\nThere are three ways to configure Instructor to use Anthropic.\n\n1. via `Mix.install([...], [instructor: [adapter: Instructor.Adapters.Anthropic, anthropic: [...]]])`\n2. via `config :instructor, adapter: Instructor.Adapters.Anthropic, anthropic: [...]`\n3. At runtime via `Instructor.chat_completion(..., config)`\n\nFor brevity, in this livebook, we'll configure it at runtime\n\n```elixir\nconfig = [\n  adapter: Instructor.Adapters.Anthropic,\n  api_key: System.fetch_env!(\"LB_ANTHROPIC_API_KEY\")\n]\n\ndefmodule President do\n  use Ecto.Schema\n\n  @primary_key false\n  embedded_schema do\n    field(:first_name, :string)\n    field(:last_name, :string)\n    field(:entered_office_date, :date)\n  end\nend\n\nInstructor.chat_completion(\n  [\n    model: \"claude-3-5-haiku-20241022\",\n    mode: :tools,\n    max_tokens: 1000,\n    response_model: President,\n    messages: [\n      %{role: \"user\", content: \"Who was the first president of the United States?\"}\n    ]\n  ],\n  config\n)\n```\n\nTo get a list of all the models supported by Anthropic, you can check out this [link](https://docs.anthropic.com/en/docs/about-claude/models).\n\n<!-- livebook:{\"offset\":1557,\"stamp\":{\"token\":\"XCP.-UxhDlnhQ0DS77yxvdRwt5881Nw7zCilkTFulcE2ia2KjbsHns2y03cJ2CNQq4-7Zerxt3tYBWGoidkVSHtHJ8e5H5fyWzIzb-CdP-gqOg8STvSaoJnE5sz_KvY\",\"version\":2}} -->","ref":"anthropic.html#introduction"},{"type":"extras","title":"Groq","doc":"# Groq\n\n```elixir\nMix.install(\n  [\n    {:instructor, path: Path.expand(\"../../\", __DIR__)},\n  ]\n)\n```","ref":"groq.html"},{"type":"extras","title":"Introduction - Groq","doc":"[Groq](https://groq.com/) is an LLM provider that uses custom chips to provide considerably higher tokens/sec for open source models than your standard AI labs using Nvidia hardware. That being said they currently only run OSS models and they tend to lag in capability from the frontier models by 6-8 months. Groq is a great option if you want speed and good price.\n\nTo get started make an account with Groq and get an API key.\n\nThere are three ways to configure Instructor to use Groq.\n\n1. via `Mix.install([...], [instructor: [adapter: Instructor.Adapters.Groq, groq: [...]]])`\n2. via `config :instructor, adapter: Instructor.Adapters.Groq, groq: [...]`\n3. At runtime via `Instructor.chat_completion(..., config)`\n\nFor brevity, in this livebook, we'll configure it at runtime.\n\n```elixir\nconfig = [\n  adapter: Instructor.Adapters.Groq,\n  api_key: System.fetch_env!(\"LB_GROQ_API_KEY\"),\n]\n\ndefmodule President do\n  use Ecto.Schema\n\n  @primary_key false\n  embedded_schema do\n    field(:first_name, :string)\n    field(:last_name, :string)\n    field(:entered_office_date, :date)\n  end\nend\n\nInstructor.chat_completion(\n  [\n    model: \"llama-3.3-70b-versatile\",\n    mode: :tools,\n    response_model: President,\n    messages: [\n      %{role: \"user\", content: \"Who was the first president of the United States?\"}\n    ]\n  ],\n  config\n)\n```\n\nAnd there you go, Instructor running at 275 tokens/second.\n\n<!-- livebook:{\"offset\":1512,\"stamp\":{\"token\":\"XCP.RW5Hd0tGmpybHXleSB6YIuSjrJCiSyPaeQnNxGRbpJ9NM3Fy46yFnJNThDeCvv3vHeA00Pu04XUh0gIxt15k6Fq4j9MzEskZWyOOwZZ58ah1FEB8B8eM\",\"version\":2}} -->","ref":"groq.html#introduction"},{"type":"extras","title":"Azure OpenAI","doc":"# Azure OpenAI\n\nConfigure your project like so to [issue requests against Azure OpenAI](https://learn.microsoft.com/en-us/azure/ai-services/openai/reference#chat-completions), according to the [docs](https://learn.microsoft.com/en-us/azure/ai-services/openai/reference#authentication).\n\n```elixir\nazure_openai_endpoint = \"https://contoso.openai.azure.com\"\nazure_openai_deployment_name = \"contosodeployment123\"\nazure_openai_api_path = \"/openai/deployments/#{azure_openai_deployment_name}/chat/completions?api-version=2024-02-01\"\n```\n\nThe Azure OpenAI service supports two authentication methods, API key and Entra ID. API key-based authN is conveyed in the `api-key` HTTP header, while Entra ID-issued access tokens go into the `Authorization: Bearer` header:","ref":"azure-openai.html"},{"type":"extras","title":"API Key Authentication - Azure OpenAI","doc":"```elixir\nconfig: [\n  instructor: [\n    adapter: Instructor.Adapters.OpenAI,\n    openai: [\n      auth_mode: :api_key_header, \n      api_key: System.get_env(\"LB_AZURE_OPENAI_API_KEY\"), # e.g. \"c3829729deadbeef382938acdfee2987\"\n      api_url: azure_openai_endpoint,\n      api_path:azure_openai_api_path\n    ]\n  ]\n]\n```","ref":"azure-openai.html#api-key-authentication"},{"type":"extras","title":"Microsoft Entra ID authentication - Azure OpenAI","doc":"The code below contains a simple GenServer that continuously refreshes the access token for a service principal. Instead of setting the configuration to a fixed access token (that would expire after an hour), the `api_key` is set to a /0-arity function that returns the most recently fetched access token.\n\n```elixir\ndefmodule AzureServicePrincipalTokenRefresher do\n  use GenServer\n\n  @derive {Inspect,\n           only: [:tenant_id, :client_id, :scope, :error], except: [:client_secret, :access_token]}\n  @enforce_keys [:tenant_id, :client_id, :client_secret, :scope]\n  defstruct [:tenant_id, :client_id, :client_secret, :scope, :access_token, :error]\n\n  def get_token_func!(tenant_id, client_id, client_secret, scope) do\n    {:ok, pid} = __MODULE__.start_link(tenant_id, client_id, client_secret, scope)\n\n    fn ->\n      case __MODULE__.get_access_token(pid) do\n        {:ok, access_token} -> access_token\n        {:error, error} -> raise \"Could not fetch Microsoft Entra ID token: #{inspect(error)}\"\n      end\n    end\n  end\n\n  def start_link(tenant_id, client_id, client_secret, scope) do\n    GenServer.start_link(__MODULE__, %__MODULE__{\n      tenant_id: tenant_id,\n      client_id: client_id,\n      client_secret: client_secret,\n      scope: scope\n    })\n  end\n\n  def get_access_token(pid) do\n    GenServer.call(pid, :get_access_token)\n  end\n\n  @impl GenServer\n  def init(%__MODULE__{} = state) do\n    {:ok, state, {:continue, :fetch_token}}\n  end\n\n  @impl GenServer\n  def handle_call(:get_access_token, _from, %__MODULE__{} = state) do\n    case state do\n      %__MODULE__{access_token: access_token, error: nil} ->\n        {:reply, {:ok, access_token}, state}\n\n      %__MODULE__{access_token: nil, error: error} ->\n        {:reply, {:error, error}, state}\n    end\n  end\n\n  @impl GenServer\n  def handle_continue(:fetch_token, %__MODULE__{} = state) do\n    {:noreply, fetch_token(state)}\n  end\n\n  @impl GenServer\n  def handle_info(:refresh_token, %__MODULE__{} = state) do\n    {:noreply, fetch_token(state)}\n  end\n\n  defp fetch_token(%__MODULE__{} = state) do\n    %__MODULE__{\n      tenant_id: tenant_id,\n      client_id: client_id,\n      client_secret: client_secret,\n      scope: scope\n    } = state\n\n    case Req.post(\n           url: \"https://login.microsoftonline.com/#{tenant_id}/oauth2/v2.0/token\",\n           form: [\n             grant_type: \"client_credentials\",\n             scope: scope,\n             client_id: client_id,\n             client_secret: client_secret\n           ]\n         ) do\n      {:ok,\n       %Req.Response{\n         status: 200,\n         body: %{\n           \"access_token\" => access_token,\n           \"expires_in\" => expires_in\n         }\n       }} ->\n        fetch_new_token_timeout = to_timeout(%Duration{second: expires_in - 60})\n        Process.send_after(self(), :refresh_token, fetch_new_token_timeout)\n        %__MODULE__{state | access_token: access_token, error: nil}\n\n      {:ok, response} ->\n        %__MODULE__{state | access_token: nil, error: response}\n\n      {:error, error} ->\n        %__MODULE__{state | access_token: nil, error: error}\n    end\n  end\nend\n```\n\nThen use the helper class to configure the dynamic credential:\n\n```elixir\nconfig: [\n  instructor: [\n    adapter: Instructor.Adapters.OpenAI,\n    openai: [\n      auth_mode: :bearer, \n      api_key: AzureServicePrincipalTokenRefresher.get_token_func!(\n          System.get_env(\"LB_AZURE_ENTRA_TENANT_ID\"), # e.g. \"contoso.onmicrosoft.com\"\n          System.get_env(\"LB_AZURE_OPENAI_CLIENT_ID\"), # e.g. \"deadbeef-0000-4f13-afa9-c8a1e4087f97\"\n          System.get_env(\"LB_AZURE_OPENAI_CLIENT_SECRET\"), # e.g. \"mEf8Q~.e2e8URInwinsermNe8wDewsedRitsen..\"}, \n          \"https://cognitiveservices.azure.com/.default\"\n      ),  \n      api_url: azure_openai_endpoint,\n      api_path: azure_openai_api_path\n    ]\n  ]\n]\n```","ref":"azure-openai.html#microsoft-entra-id-authentication"},{"type":"extras","title":"Text Classification","doc":"<!-- livebook:{\"persist_outputs\":true} -->\n\n# Text Classification\n\n```elixir\nMix.install(\n  [\n    {:instructor, path: Path.expand(\"../../\", __DIR__)}\n  ],\n  config: [\n    instructor: [\n      adapter: Instructor.Adapters.OpenAI,\n      openai: [api_key: System.fetch_env!(\"LB_OPENAI_API_KEY\")]\n    ]\n  ]\n)\n```","ref":"text-classification.html"},{"type":"extras","title":"Motivation - Text Classification","doc":"Text classification is a common task in NLP and broadly applicable across software. Whether it be spam detection, or support ticket categorization, NLP is at the core. Historically, this required training custom, bespoke models that required collecting thousands of pre-labeled examples. With LLMs a lot of this knowledge is already encoded into the model. With proper instruction and guiding the output to a known set of classifications using GPT you can be up and running with a text classification model in no time.\n\nHell, you can even use instructor to help generate the training set to train your own more efficient model. But let's not get ahead of ourselves, there's more on that later in the tutorials.","ref":"text-classification.html#motivation"},{"type":"extras","title":"Binary Text Classification - Text Classification","doc":"Spam detection is a classic example of binary text classification. It's as simple as returning a true / false of whether an example is in the class. This is pretty trivial to implement in instructor.\n\n````elixir\ndefmodule SpamPrediction do\n  use Ecto.Schema\n  use Instructor\n\n  @llm_doc \"\"\"","ref":"text-classification.html#binary-text-classification"},{"type":"extras","title":"Field Descriptions: - Text Classification","doc":"- class: Whether or not the email is spam.\n  - reason: A short, less than 10 word rationalization for the classification.\n  - score: A confidence score between 0.0 and 1.0 for the classification.\n  \"\"\"\n  @primary_key false\n  embedded_schema do\n    field(:class, Ecto.Enum, values: [:spam, :not_spam])\n    field(:reason, :string)\n    field(:score, :float)\n  end\n\n  @impl true\n  def validate_changeset(changeset) do\n    changeset\n    |> Ecto.Changeset.validate_number(:score,\n      greater_than_or_equal_to: 0.0,\n      less_than_or_equal_to: 1.0\n    )\n  end\nend\n\nis_spam? = fn text ->\n  Instructor.chat_completion(\n    model: \"gpt-4o-mini\",\n    response_model: SpamPrediction,\n    max_retries: 3,\n    messages: [\n      %{\n        role: \"user\",\n        content: \"\"\"\n        Your purpose is to classify customer support emails as either spam or not.\n        This is for a clothing retail business.\n        They sell all types of clothing.\n\n        Classify the following email: \n        ```\n        #{text}\n        ```\n        \"\"\"\n      }\n    ]\n  )\nend\n\nis_spam?.(\"Hello I am a Nigerian prince and I would like to send you money\")\n````\n\n<!-- livebook:{\"output\":true} -->\n\n```\n{:ok, %SpamPrediction{class: :spam, reason: \"Nigerian prince scam email\", score: 0.95}}\n```\n\nWe don't have to stop just at a boolean inclusion, we can also easily extend this idea to multiple categories or classes that we can classify the text into. In this example, let's consider classifying support emails. We want to know whether it's a `general_inquiry`, `billing_issue`, or a `technical_issue` perhaps it rightly fits in multiple classes. This can be useful if we want to cc' specialized support agents when intersecting customer issues occur\n\nWe can leverage `Ecto.Enum` to define a schema that restricts the LLM output to be a list of those values. We can also provide a `@doc` description to help guide the LLM with the semantic understanding of what these classifications ought to represent.\n\n```elixir\ndefmodule EmailClassifications do\n  use Ecto.Schema\n  use Instructor\n\n  @llm_doc \"\"\"\n  A classification of a customer support email.\n\n  technical_issue - whether the user is having trouble accessing their account.\n  billing_issue - whether the customer is having trouble managing their billing or credit card\n  general_inquiry - all other issues\n  \"\"\"\n  @primary_key false\n  embedded_schema do\n    field(:tags, {:array, Ecto.Enum},\n      values: [:general_inquiry, :billing_issue, :technical_issue]\n    )\n  end\nend\n\nclassify_email = fn text ->\n  {:ok, %{tags: result}} =\n    Instructor.chat_completion(\n      model: \"gpt-4o-mini\",\n      response_model: EmailClassifications,\n      messages: [\n        %{\n          role: \"user\",\n          content: \"Classify the following text: #{text}\"\n        }\n      ]\n    )\n\n  result\nend\n\nclassify_email.(\"My account is locked and I can't access my billing info.\")\n```\n\n<!-- livebook:{\"output\":true} -->\n\n```\n[:technical_issue, :billing_issue]\n```\n\n<!-- livebook:{\"offset\":4354,\"stamp\":{\"token\":\"XCP.3uvJS3qVpMhOA0gel6G7S4HKW92RcVP5l47QP8qFOhr6s9pMScBaNIBVOmf2NQKPdQoINMVH8qQKSeXVaS6gNoZDlQ6mvowO6PqbEoOX6CS00iJxeyeBqfU\",\"version\":2}} -->","ref":"text-classification.html#field-descriptions"},{"type":"extras","title":"Question & Answer with Citations","doc":"<!-- livebook:{\"persist_outputs\":true} -->\n\n# Question & Answer with Citations\n\n```elixir\nMix.install(\n  [\n    {:instructor, path: Path.expand(\"../../\", __DIR__)}\n  ],\n  config: [\n    instructor: [\n      adapter: Instructor.Adapters.OpenAI,\n      openai: [api_key: System.fetch_env!(\"LB_OPENAI_API_KEY\")]\n    ]\n  ]\n)\n```","ref":"qa-citations.html"},{"type":"extras","title":"Motivation - Question & Answer with Citations","doc":"Hallucinations are a concern with large language models. You can often ask questions and get back seemingly correct responses, but when you actually try to do quality control, you'll notice that the results are false. They just sound plausible.\n\nOne way to mitigate these shortfalls in the technology is to get the language model to provide evidence and citations to back up its answers. You can then use standard ecto changeset validation techniques to ensure that the citations also show up in the answer and the source text.","ref":"qa-citations.html#motivation"},{"type":"extras","title":"Basic Substring Validation - Question & Answer with Citations","doc":"In the simplest example of this, we can just write a validation that ensures the citation provided is found in the provided text.\n\nWe can do this pretty easily with instructor because the `validate_changeset/2` callback is optionally provided a context which you can specify for each completion.\n\n```elixir\ndefmodule QuestionAnswer do\n  use Ecto.Schema\n  use Instructor\n\n  @llm_doc \"\"\"\n  A question/answer pair with full citations from the provided supporting text.\n  citations should include the relevant facts related to answering the question\n  and quotes from the supplied text.\n  \"\"\"\n  @primary_key false\n  embedded_schema do\n    field(:question, :string)\n\n    embeds_many :answer, Citation, primary_key: false do\n      field(:statement, :string)\n      field(:quote, :string)\n    end\n  end\n\n  @impl true\n  @doc \"\"\"\n  Notice here we're getting a context? \n  We can define this value later in the `Instructor.chat_completion/1` call\n  \"\"\"\n  def validate_changeset(changeset, %{document: _document} = context) do\n    changeset\n    |> Ecto.Changeset.cast_embed(:answer,\n      with: fn params, attrs ->\n        validate_answer(context, params, attrs)\n      end\n    )\n  end\n\n  def validate_answer(%{document: document}, params, attrs \\\\ %{}) do\n    params\n    |> Ecto.Changeset.cast(attrs, [:statement, :quote])\n    |> Ecto.Changeset.validate_length(:quote, min: 1)\n    |> Ecto.Changeset.validate_change(:quote, fn :quote, q ->\n      if String.contains?(document, q) do\n        []\n      else\n        [quote: \"Quote must be an exact substring of the provided text\"]\n      end\n    end)\n  end\nend\n```\n\n<!-- livebook:{\"output\":true} -->\n\n```\n{:module, QuestionAnswer, <<70, 79, 82, 49, 0, 0, 55, ...>>, {:validate_answer, 3}}\n```\n\nNotice something interesting about this code. While we provide a validation for the change set, Instructor by default will also cast all of the embedded and associated schemas. We can override that validation of associated schemas by using the `Ecto.Changeset.cast_embed/3` function with the `with:` option.\n\nThis unfortunately won't automatically cast its attributes, but at this point you're in familiar territory with what you should be used to operating in with plain old Ecto schemas.\n\nIf at this point you don't want to fully eject from instructor, you could instead of using `Ecto.Changeset.cast_embed`, you could use `Ecto.Changeset.validate_changes(:answer, ...)` and validate the embedded changesets manually there. Both are perfectly fine.\n\nNow let's see how it responds to a question about me. Referencing my personal website.\n\n```elixir\nanswer_with_citations = fn question, context ->\n  Instructor.chat_completion(\n    model: \"gpt-4o-mini\",\n    response_model: QuestionAnswer,\n    max_retries: 2,\n    validation_context: %{\n      document: context\n    },\n    messages: [\n      %{\n        role: \"system\",\n        content:\n          \"You are a world class algorithm to answer questions with correct and exact citations.\"\n      },\n      %{\n        role: \"user\",\n        content: \"#{context}\"\n      },\n      %{\n        role: \"user\",\n        content: \"Question: #{question}\"\n      }\n    ]\n  )\nend\n\nquestion = \"What companies and what side projects has the author worked on?\"\n\ncontext = \"\"\"\n[excerpt from Thomas Millar's personal website]\n\nI have been a software engineer for the last 14 years.\nOver that time I have worked at Mortar Data (acquired by DataDog), and Stitch Fix building data platforms and doing MLOps.\nDuring these years at Stitch Fix, and in University I became close friends with Jason Liu the original author of the Python instructor library.\n\nOn the side, I have built projects like billclintonswag.com and 12ft.io which cumulatively have reached over 50M+ people directly.\nMy project's have even been written up in Oprah Magazine.\n\nThese days, I'm all in on Elixir.\nI think it's poised to be the most productive stack for SaaS companies going forward.\nI am focusing my time on building products using Elixir and developing the surrounding ecosystem.\nIf you'd like to follow me on this journey.\nCheck back here for regular updates on Elixir and general thoughts about software engineering.\nYou can also follow me on  @thmsmlr.\n\"\"\"\n\nanswer_with_citations.(question, context)\n```\n\n<!-- livebook:{\"output\":true} -->\n\n```\n{:ok,\n %QuestionAnswer{\n   question: \"What companies and what side projects has the author worked on?\",\n   answer: [\n     %QuestionAnswer.Citation{\n       statement: \"The author has worked for Mortar Data and Stitch Fix.\",\n       quote: \"I have worked at Mortar Data (acquired by DataDog), and Stitch Fix building data platforms and doing MLOps.\"\n     },\n     %QuestionAnswer.Citation{\n       statement: \"The author has built side projects like billclintonswag.com and 12ft.io.\",\n       quote: \"On the side, I have built projects like billclintonswag.com and 12ft.io which cumulatively have reached over 50M+ people directly.\"\n     },\n     %QuestionAnswer.Citation{\n       statement: \"The author's projects have been featured in Oprah Magazine.\",\n       quote: \"My project's have even been written up in Oprah Magazine.\"\n     }\n   ]\n }}\n```","ref":"qa-citations.html#basic-substring-validation"},{"type":"extras","title":"Using the LLM for Validation - Question & Answer with Citations","doc":"While this certainly works, you don't really have a strong confidence that the citation is actually relevant, even if it is found in the base text. However, the beauty of Instructor is that you can employ it recursively in your validators to check its own work.\n\nSo in this example, we're going to write a custom validator for the citations that check against the base text with the LLM to ensure that it is semantically relevant, not just present.\n\n```elixir\ndefmodule CitationValidation do\n  use Ecto.Schema\n  use Instructor\n\n  @llm_doc \"\"\"\n  Whether or not a citation for a given text is valid.\n  Optionally you can provide an error_message to when the citation is invalid.\n  \"\"\"\n  @primary_key false\n  embedded_schema do\n    field(:is_valid?, :boolean)\n    field(:error_message, :string)\n  end\n\n  def changeset(params, attrs \\\\ %{}) do\n    params\n    |> Ecto.Changeset.cast(attrs, [:is_valid?, :error_message])\n  end\nend\n\ndefmodule LLMQuestionAnswer do\n  use Ecto.Schema\n  use Instructor\n\n  @llm_doc \"\"\"\n  A question/answer pair with full citations from the provided supporting text.\n  citations should include the relevant facts related to answering the question\n  and quotes from the supplied text.\n  \"\"\"\n  @primary_key false\n  embedded_schema do\n    field(:question, :string)\n\n    embeds_many :answer, Citation, primary_key: false do\n      field(:statement, :string)\n      field(:quote, :string)\n    end\n  end\n\n  @impl true\n  def validate_changeset(changeset, %{document: _document} = context) do\n    changeset\n    |> Ecto.Changeset.cast_embed(:answer,\n      with: fn params, attrs ->\n        validate_answer(context, params, attrs)\n      end\n    )\n  end\n\n  def validate_answer(%{document: document}, params, attrs \\\\ %{}) do\n    params\n    |> Ecto.Changeset.cast(attrs, [:statement, :quote])\n    |> Ecto.Changeset.validate_length(:quote, min: 1)\n    |> Ecto.Changeset.validate_change(:quote, fn :quote, q ->\n      case Instructor.chat_completion(\n             model: \"gpt-4o-mini\",\n             response_model: CitationValidation,\n             max_retries: 3,\n             messages: [\n               %{\n                 role: \"user\",\n                 content: \"\"\"\n                   Does the following citation exist in the following document?\n                   It is okay if the citation is slightly wrong, but semantically correct.\n\n                   Citation: #{q}\n\n                   Context: #{document}\n                 \"\"\"\n               }\n             ]\n           ) do\n        {:ok, %CitationValidation{is_valid?: true}} -> []\n        {:ok, %CitationValidation{is_valid?: false, error_message: err}} -> [quote: err]\n      end\n    end)\n  end\nend\n```\n\n<!-- livebook:{\"output\":true} -->\n\n```\n{:module, LLMQuestionAnswer, <<70, 79, 82, 49, 0, 0, 58, ...>>, {:validate_answer, 3}}\n```\n\nNotice here that we relaxed the requirement that it must exactly be in the base text. Rather the citation can just be semantically represented. This allows us to be a little more fuzzy with our validations.\n\nLet's induce a hypothetical invalid changeset so that we can test our validations.\n\n```elixir\n# Simulated response from the LLM\nparams = %{\n  question: \"What is the capital of France?\",\n  answer: [\n    %{\n      statement: \"Paris\",\n      quote: \"Paris is the capital of France\"\n    }\n  ]\n}\n\n# Internally Instructor makes this call with the result of the LLM to create\n# the changeset that it'll later validate.\nchangeset = Instructor.cast_all(%LLMQuestionAnswer{}, params)\n\n%Ecto.Changeset{valid?: true} =\n  LLMQuestionAnswer.validate_changeset(changeset, %{\n    document: \"\"\"\n      Thomas likes to golf\n      Paris is the capital city of France\n      Some other irrelevant text\n    \"\"\"\n  })\n```\n\n<!-- livebook:{\"output\":true} -->\n\n```\n#Ecto.Changeset ,\n        valid?: true,\n        ...\n      >\n    ]\n  },\n  errors: [],\n  data: #LLMQuestionAnswer<>,\n  valid?: true,\n  ...\n>\n```\n\n```elixir\n# Or and example of the citation failing\n\n# Simulated response from the LLM\nparams = %{\n  question: \"What is the capital of France?\",\n  answer: [\n    %{\n      statement: \"Paris\",\n      quote: \"Paris is the capital of France\"\n    }\n  ]\n}\n\n# Internally Instructor makes this call with the result of the LLM to create\n# the changeset that it'll later validate.\nchangeset = Instructor.cast_all(%LLMQuestionAnswer{}, params)\n\n%Ecto.Changeset{valid?: false} =\n  LLMQuestionAnswer.validate_changeset(changeset, %{\n    document: \"\"\"\n      Thomas likes to golf\n      Some other irrelevant text\n    \"\"\"\n  })\n```\n\n<!-- livebook:{\"output\":true} -->\n\n```\n#Ecto.Changeset ,\n        valid?: false,\n        ...\n      >\n    ]\n  },\n  errors: [],\n  data: #LLMQuestionAnswer<>,\n  valid?: false,\n  ...\n>\n```\n\n<!-- livebook:{\"offset\":11362,\"stamp\":{\"token\":\"XCP.J32qU39xdS7_7Kgs0pk5U8oDkmilSFrLb8uq9jn-ui_W2BXnB6sReBCAc2z0FMvLzq70FvJXev2GAnQrvJsq5H_dilIWvjeoyTbKHG5a3OV2CNn9RDWWugM\",\"version\":2}} -->","ref":"qa-citations.html#using-the-llm-for-validation"},{"type":"extras","title":"Extracting Action Items from Meeting Transcripts","doc":"<!-- livebook:{\"persist_outputs\":true} -->\n\n# Extracting Action Items from Meeting Transcripts\n\n```elixir\nMix.install(\n  [\n    {:instructor, path: Path.expand(\"../../\", __DIR__)},\n    {:kino, \"~> 0.12.0\"}\n  ],\n  config: [\n    instructor: [\n      adapter: Instructor.Adapters.OpenAI,\n      openai: [api_key: System.fetch_env!(\"LB_OPENAI_API_KEY\")]\n    ]\n  ]\n)\n```","ref":"extract-action-items-from-meeting-transcripts.html"},{"type":"extras","title":"Motivation - Extracting Action Items from Meeting Transcripts","doc":"This example shouldn't be foreign to any of you. I'm sure each day you log in, join a Zoom call, and have a meeting about what y'all are going to do this upcoming week. It is then the work of some product manager to translate this into a JIRA board so that you can track your progress throughout the week.\n\nThe bane of most engineers existence... We can automate this.","ref":"extract-action-items-from-meeting-transcripts.html#motivation"},{"type":"extras","title":"The Schema - Extracting Action Items from Meeting Transcripts","doc":"Let's start by defining a schema for the tickets and the subtasks that might exist within them. There will be a priority, description and a set of dependencies between the tasks for the tickets.\n\n```elixir\ndefmodule MeetingNotes do\n  use Ecto.Schema\n  use Instructor.Validator\n\n  @llm_doc \"\"\"\n  Tickets correctly resolved from a meeting transcription.\n\n  Tickets have a name, priority, useful description, and assignees.\n  They may also have subtasks that share the same id spaces as the tickets.\n  A ticket also may have references to Tickets and Subtasks that are a blocking dependency\n  for completing the ticket.\n  \"\"\"\n  @primary_key false\n  embedded_schema do\n    embeds_many :tickets, Ticket do\n      field(:name, :string)\n      field(:description, :string)\n      field(:priority, Ecto.Enum, values: [:high, :medium, :low])\n      field(:assignees, {:array, :string})\n\n      embeds_many :subtasks, SubTasks do\n        field(:name, :string)\n      end\n\n      field(:dependencies, {:array, :binary_id})\n    end\n  end\nend\n```\n\n<!-- livebook:{\"output\":true} -->\n\n```\nwarning: module attribute @llm_doc was set but never used\n└─ pages/cookbook/extract-action-items-from-meeting-transcripts.livemd#cell:2eli2kgfgvm2ligb:5: MeetingNotes (module)\n\n```\n\n<!-- livebook:{\"output\":true} -->\n\n```\n{:module, MeetingNotes, <<70, 79, 82, 49, 0, 0, 41, ...>>, :ok}\n```\n\nOne thing interesting to note here is that although we could use proptorecto associations, in this case it's rather tedious. Since our dependencies can either be on tickets or on subtasks, you would typically have to have a polymorphic relationship.\n\nWhile Instructor fully supports doing that, we can instead just create embedded schemas for the subtasks and define that the subtasks and the tickets share the same ID space in our doc comment. This will steer the LLM to produce foreign key relations in a polymorphic way.\n\nOne also might choose to solve this modeling problem by denormalizing the tickets and subtasks into its own task type and get the LLM to not only have the dependencies listed but also any subtasks that a ticket might have through association instead of embedding.\n\nEither way works, but in this example we prefer the embedded method because it produces a prettier output and we can lean on the LLM a little more.\n\nNow, let's extract some JIRA tickets.\n\n```elixir\ngenerate_tickets = fn transcript ->\n  Instructor.chat_completion(\n    model: \"gpt-4o-mini\",\n    response_model: MeetingNotes,\n    messages: [\n      %{role: \"system\", content: \"The following is a transcript of a meeting...\"},\n      %{\n        role: \"user\",\n        content: \"Create the action items for the following transcript: #{transcript}\"\n      }\n    ]\n  )\nend\n\n{:ok, %{tickets: tickets}} =\n  generate_tickets.(\"\"\"\n    Alice: Hey team, we have several critical tasks we need to tackle for the upcoming release. First, we need to work on improving the authentication system. It's a top priority.\n    \n    Bob: Got it, Alice. I can take the lead on the authentication improvements. Are there any specific areas you want me to focus on?\n    \n    Alice: Good question, Bob. We need both a front-end revamp and back-end optimization. So basically, two sub-tasks.\n    \n    Carol: I can help with the front-end part of the authentication system.\n    \n    Bob: Great, Carol. I'll handle the back-end optimization then.\n    \n    Alice: Perfect. Now, after the authentication system is improved, we have to integrate it with our new billing system. That's a medium priority task.\n    \n    Carol: Is the new billing system already in place?\n    \n    Alice: No, it's actually another task. So it's a dependency for the integration task. Bob, can you also handle the billing system?\n    \n    Bob: Sure, but I'll need to complete the back-end optimization of the authentication system first, so it's dependent on that.\n    \n    Alice: Understood. Lastly, we also need to update our user documentation to reflect all these changes. It's a low-priority task but still important.\n    \n    Carol: I can take that on once the front-end changes for the authentication system are done. So, it would be dependent on that.\n    \n    Alice: Sounds like a plan. Let's get these tasks modeled out and get started.\n  \"\"\")\n```\n\n<!-- livebook:{\"output\":true} -->\n\n```\n{:ok,\n %MeetingNotes{\n   tickets: [\n     %MeetingNotes.Ticket{\n       id: \"1\",\n       name: \"Improve authentication system\",\n       description: \"Work on improving the authentication system\",\n       priority: :high,\n       assignees: [\"Bob\"],\n       subtasks: [\n         %MeetingNotes.Ticket.SubTasks{id: \"1\", name: \"Front-end revamp\"},\n         %MeetingNotes.Ticket.SubTasks{id: \"2\", name: \"Back-end optimization\"}\n       ],\n       dependencies: []\n     },\n     %MeetingNotes.Ticket{\n       id: \"2\",\n       name: \"Integrate authentication system with new billing system\",\n       description: \"Integrate the improved authentication system with the new billing system\",\n       priority: :medium,\n       assignees: [\"Bob\"],\n       subtasks: [],\n       dependencies: [\"1\"]\n     },\n     %MeetingNotes.Ticket{\n       id: \"3\",\n       name: \"Update user documentation\",\n       description: \"Update user documentation to reflect changes\",\n       priority: :low,\n       assignees: [\"Carol\"],\n       subtasks: [],\n       dependencies: [\"1.1\"]\n     }\n   ]\n }}\n```\n\nThe results look good. We can use Kino to then render out all the dependencies of our tickets.\n\n```elixir\ngenerate_tickets_diagram = fn tickets ->\n  subtasks = Enum.flat_map(tickets, & &1.subtasks)\n  all_tasks = subtasks ++ tickets\n\n  ticket_nodes =\n    tickets\n    |> Enum.map_join(\"\\n\", fn t ->\n      subtask_relations =\n        t.subtasks\n        |> Enum.map_join(\"\\n\", fn st ->\n          \"\"\"\n          \"#{t.name}\" ||--o| \"#{st.name}\" : \"Has Subtask\"\n          \"\"\"\n        end)\n\n      dependency_relations =\n        t.dependencies\n        |> Enum.map_join(\"\\n\", fn d ->\n          dt = all_tasks |> Enum.find(&(&1.id == d))\n\n          if dt do\n            \"\"\"\n            \"#{t.name}\" ||--o| \"#{dt.name}\" : \"Depends on\"\n            \"\"\"\n          else\n            \"\"\n          end\n        end)\n\n      \"\"\"\n        \"#{t.name}\" {\n          priority #{t.priority}\n          assignees #{Enum.join(t.assignees, \", \")}\n        }\n\n        #{subtask_relations}\n        #{dependency_relations}\n      \"\"\"\n    end)\n\n  subtask_nodes =\n    tickets\n    |> Enum.flat_map(& &1.subtasks)\n    |> Enum.map_join(\"\\n\", fn st ->\n      \"\"\"\n        \"#{st.name}\"\n      \"\"\"\n    end)\n\n  Kino.Mermaid.new(\"\"\"\n    erDiagram\n      #{ticket_nodes}\n      #{subtask_nodes}\n  \"\"\")\nend\n\ngenerate_tickets_diagram.(tickets)\n```\n\n<!-- livebook:{\"output\":true} -->\n\n```mermaid\n  erDiagram\n      \"Improve authentication system\" {\n    priority high\n    assignees Bob\n  }\n\n  \"Improve authentication system\" ||--o| \"Front-end revamp\" : \"Has Subtask\"\n\n\"Improve authentication system\" ||--o| \"Back-end optimization\" : \"Has Subtask\"\n\n  \n\n  \"Integrate authentication system with new billing system\" {\n    priority medium\n    assignees Bob\n  }\n\n  \n  \"Integrate authentication system with new billing system\" ||--o| \"Front-end revamp\" : \"Depends on\"\n\n\n  \"Update user documentation\" {\n    priority low\n    assignees Carol\n  }\n\n  \n  \n\n      \"Front-end revamp\"\n\n  \"Back-end optimization\"\n\n\n```\n\n<!-- livebook:{\"offset\":8052,\"stamp\":{\"token\":\"XCP.MbYCNKWrbskKRZ0OgP2c22JFIfPb-VNrPezIkhYp8jwE9IRwwqmTDySRVJmKMLHnDXUxnbyVbA2c4_6xLzJMb_vFkiVItnA6ajI06Au4GUN3SgVokNpHvOc\",\"version\":2}} -->","ref":"extract-action-items-from-meeting-transcripts.html#the-schema"},{"type":"extras","title":"Converting Text into Explorer DataFrames","doc":"<!-- livebook:{\"persist_outputs\":true} -->\n\n# Converting Text into Explorer DataFrames\n\n```elixir\nMix.install(\n  [\n    {:instructor, path: Path.expand(\"../../\", __DIR__)},\n    {:explorer, \"~> 0.7.2\"},\n    {:kino, \"~> 0.12.0\"},\n    {:kino_explorer, \"~> 0.1.13\"}\n  ],\n  config: [\n    instructor: [\n      adapter: Instructor.Adapters.OpenAI,\n      openai: [api_key: System.fetch_env!(\"LB_OPENAI_API_KEY\")]\n    ]\n  ]\n)\n```","ref":"text-to-dataframes.html"},{"type":"extras","title":"Motivation - Converting Text into Explorer DataFrames","doc":"Sometimes we come across text that can be naturally extracted into structured data. Although we don't necessarily know how many tables, and what the schemas are represented in the data. Using Instructor we can let the LLM determine the tables and schemas for us, pulling the results into a dataframe which we can then use for further analysis.","ref":"text-to-dataframes.html#motivation"},{"type":"extras","title":"The Schema - Converting Text into Explorer DataFrames","doc":"Notice from this example we want resulting data structure to have data frames. That's not a default Ecto type so we're gonna have to do things a little different here.\n\nLuckily, Ecto allows us to define our own custom types in which we can override how the values are casted. And Instructor provides us a behavior which tells us how to represent it in a JSONSchema so the LLM can understand it.\n\nIn our case, we're going to create a type for a table of data, which will be represented as a CSV string with respect to the LLM. Our custom type will then parse the string when it is casted by Ecto and parse the CSV into an `Explorer.DataFrame`.\n\n```elixir\ndefmodule Ecto.CSVDataFrame do\n  use Ecto.Type\n  use Instructor.EctoType\n\n  def type, do: :string\n\n  def cast(csv_str) when is_binary(csv_str) do\n    df = Explorer.DataFrame.load_csv!(csv_str)\n    {:ok, df}\n  end\n\n  def cast(%Explorer.DataFrame{} = df), do: {:ok, df}\n  def cast(_), do: :error\n\n  def to_json_schema(),\n    do: %{\n      type: \"string\",\n      description: \"A CSV representation of a data table\"\n    }\n\n  def dump(x), do: {:ok, x}\n  def load(x), do: {:ok, x}\nend\n```\n\n<!-- livebook:{\"output\":true} -->\n\n```\n{:module, Ecto.CSVDataFrame, <<70, 79, 82, 49, 0, 0, 17, ...>>, {:load, 1}}\n```\n\nNow that we have our data type, we can use it in our embedded ecto schema, just as we would have in any other example. If everything works correctly, we should get some data frames out.\n\n```elixir\ndefmodule Database do\n  use Ecto.Schema\n  use Instructor\n\n  @llm_doc \"\"\"\n  The extracted database will contain one or more tables with data as csv formatted with ',' delimiters\n  \"\"\"\n  @primary_key false\n  embedded_schema do\n    embeds_many :tables, DataFrame, primary_key: false do\n      field(:name, :string)\n      field(:data, Ecto.CSVDataFrame)\n    end\n  end\nend\n```\n\n<!-- livebook:{\"output\":true} -->\n\n```\n{:module, Database, <<70, 79, 82, 49, 0, 0, 43, ...>>, :ok}\n```\n\nPerfect. And now let's test it out on some example text.\n\n```elixir\nextract_df = fn text ->\n  Instructor.chat_completion(\n    model: \"gpt-4o-mini\",\n    response_model: Database,\n    messages: [\n      %{\n        role: \"system\",\n        content: \"\"\"\n        Map this data into one or more dataframes and correctly define the columns and data\n        \"\"\"\n      },\n      %{\n        role: \"user\",\n        content: \"#{text}\"\n      }\n    ]\n  )\nend\n\n{:ok, db} =\n  extract_df.(\"\"\"\n    My name is John and I am 25 years old. I live in\n    New York and I like to play basketball. His name is\n    Mike and he is 30 years old. He lives in San Francisco\n    and he likes to play baseball. Sarah is 20 years old\n    and she lives in Los Angeles. She likes to play tennis.\n    Her name is Mary and she is 35 years old.\n    She lives in Chicago.\n    \n    On one team 'Tigers' the captain is John and there are 12 players.\n    On the other team 'Lions' the captain is Mike and there are 10 players.\n  \"\"\")\n\nKino.Layout.tabs(\n  for table <- db.tables do\n    {table.name, table.data}\n  end\n)\n```\n\n<!-- livebook:{\"offset\":3802,\"stamp\":{\"token\":\"XCP.8kFgMDCVN4NG0bG_VpKXVUz9s7AeP6rAbpnYpK8YhFFWZB1PNWkQrhphICFfRa9zH2HfwULNqz-PRLtaKjrpR9WuhDbFFLfjHsl1crH0Iyf5BEALWenu9-A\",\"version\":2}} -->","ref":"text-to-dataframes.html#the-schema"},{"type":"extras","title":"GPT-4o-vision - Extracting Data from Images","doc":"<!-- livebook:{\"file_entries\":[{\"name\":\"shopify-screenshot.png\",\"type\":\"attachment\"}],\"persist_outputs\":true} -->\n\n# GPT-4o-vision - Extracting Data from Images\n\n```elixir\nMix.install(\n  [\n    {:instructor, path: Path.expand(\"../../\", __DIR__)},\n    {:kino, \"~> 0.12.3\"}\n  ],\n  config: [\n    instructor: [\n      adapter: Instructor.Adapters.OpenAI,\n      openai: [\n        api_key: System.fetch_env!(\"LB_OPENAI_API_KEY\"),\n      ]\n    ]\n  ]\n)\n```","ref":"gpt4o-vision.html"},{"type":"extras","title":"Motivation - GPT-4o-vision - Extracting Data from Images","doc":"The latest models support vision capabilities as well. This, with no extra work, is a feature of Instructor. All you have to do is pass a URL or Base64 encoded image as one of the messages, and everything should just work seamlessly.\n\nIn the following example, we will extract product details from a screenshot of a Shopify store.\n\n<!-- livebook:{\"break_markdown\":true} -->\n\n![](files/shopify-screenshot.png)\n\n```elixir\nimage = Kino.FS.file_path(\"shopify-screenshot.png\") |> File.read!()\nbase64_image = \"data:image/png;base64,\" <> Base.encode64(image)\n\ndefmodule Product do\n  use Ecto.Schema\n\n  @primary_key false\n  embedded_schema do\n    field(:name, :string)\n    field(:price, :decimal)\n    field(:currency, Ecto.Enum, values: [:usd, :gbp, :eur, :cny])\n    field(:color, :string)\n  end\nend\n\n{:ok, result} =\n  Instructor.chat_completion(\n    model: \"gpt-4o\",\n    response_model: Product,\n    messages: [\n      %{\n        role: \"user\",\n        content: [\n          %{type: \"text\", text: \"What is the product details of the following image?\"},\n          %{type: \"image_url\", image_url: %{url: base64_image, detail: \"high\"}}\n        ]\n      }\n    ]\n  )\n\nresult\n```\n\n<!-- livebook:{\"output\":true} -->\n\n```\n%Product{\n  name: \"Thomas Wooden Railway Thomas The Tank Engine\",\n  price: Decimal.new(\"33.0\"),\n  currency: :usd,\n  color: \"blue\"\n}\n```\n\n<!-- livebook:{\"offset\":1801,\"stamp\":{\"token\":\"XCP.3eUsk2eoTssLzfx2PUW1jDxQZ3yoXNoei3RlIouC-oe4lAyYJ9NZgK3tSBeSiVeyaoXRPJYHKK82NSagHJgvelAuiw2S_BAXweLfTVQx202SgBwu9o-MtxM\",\"version\":2}} -->","ref":"gpt4o-vision.html#motivation"},{"type":"extras","title":"Extracting data from images of Receipts","doc":"<!-- livebook:{\"file_entries\":[{\"name\":\"receipt.jpg\",\"type\":\"attachment\"}],\"persist_outputs\":true} -->\n\n# Extracting data from images of Receipts\n\n```elixir\nMix.install(\n  [\n    {:instructor, path: Path.expand(\"../../\", __DIR__)},\n    {:kino, \"~> 0.12.3\"}\n  ],\n  config: [\n    instructor: [\n      adapter: Instructor.Adapters.OpenAI,\n      openai: [\n        api_key: System.fetch_env!(\"LB_OPENAI_API_KEY\"),\n      ]\n    ]\n  ]\n)\n```","ref":"ocr-receipts-with-gpt4o.html"},{"type":"extras","title":"Our Object Model - Extracting data from images of Receipts","doc":"```elixir\nimage = Kino.FS.file_path(\"receipt.jpg\") |> File.read!()\n```\n\n<!-- livebook:{\"output\":true} -->\n\n```\n<<255, 216, 255, 224, 0, 16, 74, 70, 73, 70, 0, 1, 1, 0, 0, 1, 0, 1, 0, 0, 255, 226, 2, 40, 73, 67,\n  67, 95, 80, 82, 79, 70, 73, 76, 69, 0, 1, 1, 0, 0, 2, 24, 0, 0, 0, 0, 2, 16, 0, 0, ...>>\n```\n\nWe can look at the image of the receipt and outline the fields that we want to extract using an Ecto schema, as shown below.\n\n```elixir\ndefmodule Receipt do\n  use Ecto.Schema\n  use Instructor.Validator\n  \n  import Ecto.Changeset\n\n  @primary_key false\n  embedded_schema do\n    field(:total, :decimal)\n    field(:subtotal, :decimal)\n\n    embeds_many :items, Item do\n      field(:name, :string)\n      field(:price, :decimal)\n      field(:quantity, :integer)\n    end\n  end\n\n  @impl true\n  def validate_changeset(changeset) do\n    changeset\n    |> validate_required([:total, :subtotal])\n    |> validate_items_total()\n  end\n\n  defp validate_items_total(changeset) do\n    items = get_field(changeset, :items) || []\n    subtotal = get_field(changeset, :subtotal)\n\n    items_total = Enum.reduce(items, Decimal.new(0), fn item, acc ->\n      item_total = Decimal.mult(item.price, Decimal.new(item.quantity))\n      Decimal.add(acc, item_total)\n    end)\n\n    if Decimal.equal?(items_total, subtotal) do\n      changeset\n    else\n      add_error(changeset, :subtotal, \"\"\"\n        Subtotal does not match sum of item prices. \n        Subtotal is #{subtotal} and items total is #{items_total}\")\n      \"\"\")\n    end\n  end\nend\n\nKino.nothing()\n```","ref":"ocr-receipts-with-gpt4o.html#our-object-model"},{"type":"extras","title":"Validating the Subtotal - Extracting data from images of Receipts","doc":"Notice how we used the `validate_changeset` callback to check that the items we extract sum up to the subtotal on the receipt. This process allows us to gain confidence that the OCR application is functioning correctly. In fact, later on, we can even use the re-ask feature of the instructor to have the LLM correct any errors itself.","ref":"ocr-receipts-with-gpt4o.html#validating-the-subtotal"},{"type":"extras","title":"Calling the LLM - Extracting data from images of Receipts","doc":"```elixir\nbase64_image = \"data:image/jpeg;base64,\" <> Base.encode64(image)\n\n{:ok, receipt} = Instructor.chat_completion(\n  model: \"gpt-4o\",\n  response_model: Receipt,\n  messages: [\n    %{\n      role: \"user\",\n      content: [\n        %{\n          type: \"image_url\",\n          image_url: %{url: base64_image}\n        },\n        %{\n          type: \"text\",\n          text: \"Analyze the image and return the items in the receipt and the total amount.\"\n        }\n      ]\n    }\n  ]\n)\n\nKino.Layout.grid([Kino.Image.new(image, :jpeg), receipt], columns: 2, boxed: true)\n```\n\n<!-- livebook:{\"output\":true} -->\n\n```\n%Receipt{\n  total: Decimal.new(\"107.6\"),\n  subtotal: Decimal.new(\"107.6\"),\n  items: [\n    %Receipt.Item{id: \"1\", name: \"Lorem ipsum\", price: Decimal.new(\"9.2\"), quantity: 1},\n    %Receipt.Item{id: \"2\", name: \"Lorem ipsum dolor sit\", price: Decimal.new(\"19.2\"), quantity: 1},\n    %Receipt.Item{\n      id: \"3\",\n      name: \"Lorem ipsum dolor sit amet\",\n      price: Decimal.new(\"15.0\"),\n      quantity: 1\n    },\n    %Receipt.Item{id: \"4\", name: \"Lorem ipsum\", price: Decimal.new(\"15.0\"), quantity: 1},\n    %Receipt.Item{id: \"5\", name: \"Lorem ipsum\", price: Decimal.new(\"15.0\"), quantity: 1},\n    %Receipt.Item{id: \"6\", name: \"Lorem ipsum dolor sit\", price: Decimal.new(\"15.0\"), quantity: 1},\n    %Receipt.Item{id: \"7\", name: \"Lorem ipsum\", price: Decimal.new(\"19.2\"), quantity: 1}\n  ]\n}\n```\n\nNow, we simply using Instructor to call gpt-4o with the base64 encoded image and the response model, and we can get back the results. We can have confidence that the results returned match all of our validations, and we've reduced the effects of any hallucinations.\n\n<!-- livebook:{\"offset\":4033,\"stamp\":{\"token\":\"XCP.w3F951NvQo7WPlhcGI-eUyFZogp1tgDIZalsbErufIRG3vdlaJkCHkZKz2q86X6OqesLdco1CFtKzOz3ZmNYHUk36DExApeehftSh1NzpW4CkNtkdweK3is\",\"version\":2}} -->","ref":"ocr-receipts-with-gpt4o.html#calling-the-llm"},{"type":"extras","title":"PII Data Sanitization","doc":"# PII Data Sanitization","ref":"pii-data-sanitization.html"},{"type":"extras","title":"Section - PII Data Sanitization","doc":"```elixir\nMix.install(\n  [\n    {:instructor, path: Path.expand(\"../../\", __DIR__)}\n  ],\n  config: [\n    instructor: [\n      adapter: Instructor.Adapters.OpenAI,\n      openai: [api_key: System.fetch_env!(\"LB_OPENAI_API_KEY\")]\n    ]\n  ]\n)\n```\n\n#","ref":"pii-data-sanitization.html#section"},{"type":"extras","title":"Overview - PII Data Sanitization","doc":"This example demonstrates the usage of OpenAI's ChatCompletion model for the extraction and scrubbing of Personally Identifiable Information (PII) from an input. The code defines Ecto schema to manage the PII data and offers function for both extraction and sanitation.\n\n##","ref":"pii-data-sanitization.html#overview"},{"type":"extras","title":"Defining the Structures - PII Data Sanitization","doc":"First, Ecto schemas are defined to represent the PII data and the overall structure for PII data extraction.\n\n```elixir\ndefmodule PII do\n  use Ecto.Schema\n  use Instructor\n\n  @llm_doc \"\"\"","ref":"pii-data-sanitization.html#defining-the-structures"},{"type":"extras","title":"Field Descriptions: - PII Data Sanitization","doc":"- index: an auto incrementing integer starting at zero\n  - type: the type of personal identifiable information\n  - value: the PII value\n  \"\"\"\n  @primary_key false\n  embedded_schema do\n    embeds_many :data, Datum, primary_key: false do\n      field(:index, :integer)\n      field(:type, :string)\n      field(:value, :string)\n    end\n  end\n\n  @doc \"\"\"\n  Iterates over the private data and replaces the value with a placeholder in the\n  form of <{data_type}_{i}>\n  \"\"\"\n  def scrub({:ok, pii}, input) do\n    Enum.reduce(pii.data, input, fn datum, acc ->\n      String.replace(acc, datum.value, \"<#{datum.type}_#{datum.index}>\")\n    end)\n  end\n\n  def scrub({:error, reason}, _input) do\n    dbg(reason)\n  end\n\n  def extract(input) do\n    Instructor.chat_completion(\n      model: \"gpt-4o-mini\",\n      response_model: PII,\n      max_retries: 3,\n      messages: [\n        %{\n          role: \"system\",\n          content:\n            \"You are a world class PII scrubbing model, Extract the PII data from the following document\"\n        },\n        %{\n          role: \"system\",\n          content: \"\"\"\n          Examples of PII: names, addresses, phone numbers, email addresses, financial information\n          \"\"\"\n        },\n        %{\n          role: \"system\",\n          content: \"\"\"\n          Instructions:\n          - any spaces in the type should be converted to underscores and all letters should be lower case\n          - use abbreviations when choosing the type\n          \"\"\"\n        },\n        %{\n          role: \"user\",\n          content: input\n        }\n      ]\n    )\n  end\nend\n```\n\n#","ref":"pii-data-sanitization.html#field-descriptions"},{"type":"extras","title":"Extracting PII Data - PII Data Sanitization","doc":"The OpenAI API is utilized to extract PII information from a given input.\n\n```elixir\ninput =\n  \"Hello John Smith, I am Jill. Your GitBoat, LLC credit card account 1111-0000-1111-8765 has a minimum payment of $33.32 that is due by July 24th.\"\n\npii_data = PII.extract(input)\n```\n\n#","ref":"pii-data-sanitization.html#extracting-pii-data"},{"type":"extras","title":"Scrubbing PII Data - PII Data Sanitization","doc":"After extracting the PII data, the `PII.scrub/2` funnction is used to sanitize the input.\n\n```elixir\nPII.scrub(pii_data, input)\n```\n\n<!-- livebook:{\"offset\":2846,\"stamp\":{\"token\":\"XCP.GOftTlzRuHpa0i4xzSq4TkWzY7HJSZtYLdXs25MshgHnAhgL6AYaxz23eLYIf4JUKCoIYxL2SaoYNYz-YKVdvwJX7n9xfEmuAv5Tui_Xr8ANFkIeVlttKBo\",\"version\":2}} -->","ref":"pii-data-sanitization.html#scrubbing-pii-data"}],"content_type":"text/markdown"}